/*      file     : TrafficData/TMultilaneGrid.C
        project  : PAMINA III
        author   : Marcus Rickert
        objects  : TMultilaneGridConfig
                   TMultilaneGrid
        RCS      : $Id: TMultilaneGrid.C,v 1.14 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

/* ***INCLUDES************************************************************************ */

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <TStream.h>
#include <TTransferInfo.h>
#include <TDistributedRange.h>
#include <TTimeStep.h>
#include <TObjectManager.h>
#include <TMultilaneBoundary.h>
#include <TRouteplan.h>
#include <TVehicle.h>

#include "TMultilaneGrid.h"

// ***DEFINES*************************************************************************

#define MAX_INDEXED_ENCODE_RATIO         0.3

#define LC_INDEX_WIDTH                      24
#define LC_FROM_LANE_SHIFT     (LC_INDEX_WIDTH)
#define LC_TO_LANE_SHIFT       (LC_FROM_LANE_SHIFT+3)
#define LC_INDEX_MASK               0x00FFFFFF
#define LC_LANE_MASK                0x00000007
#define LC_GET_SITE_INDEX(Info) ((Info) & LC_INDEX_MASK)
#define LC_GET_FROM_LANE(Info) (((Info) & (LC_LANE_MASK << LC_FROM_LANE_SHIFT)) >> LC_FROM_LANE_SHIFT)
#define LC_GET_TO_LANE(Info) (((Info) & (LC_LANE_MASK << LC_TO_LANE_SHIFT)) >> LC_TO_LANE_SHIFT)
#define LC_COMPOSE(Index, FromLane, ToLane) ((Index) | ((FromLane) << LC_FROM_LANE_SHIFT) | ((ToLane) << LC_TO_LANE_SHIFT))

// define access functions for 'Sites', 'Vehicles', and their scratch counterparts

#define AllSites(CSite) (CSite = FirstSite; CSite < LastSite; CSite++)
#define AllActiveSites(CSite) (CSite = FirstActiveSite; CSite < LastActiveSite; CSite++)
#define AllTransferSites(CSite) (CSite = TransferFirstSite; CSite < TransferLastSite; CSite++)

#define PredGap(Current) (Pred-Current-1)
#define PredLGap(CLane, Current) (Pred[CLane]-Current-1)

// ***GLOBAL VARIABLES****************************************************************  

TClassInfo *TMultilaneGrid::ClassInfo = 0;

/* ***METHODS************************************************************************* */

#ifdef NO_INLINE
#include <TMultilaneGrid.i.C>
#endif

/* =================================================================================== */

#include "MicroSimCA.C"

/* =================================================================================== */

TMultilaneGridVarData::TMultilaneGridVarData(void)

{
    OccupiedSites = 0;
}

/* =================================================================================== */

bool
TMultilaneGridVarData::Decode(TBaseStream *aStream)

{
    return aStream->Decode(OccupiedSites);
}

/* =================================================================================== */

bool
TMultilaneGridVarData::Encode(TBaseStream *aStream) const

{
    return aStream->Encode(OccupiedSites);
}

/* =================================================================================== */
#define ParentClass TParallelObject
/* =================================================================================== */

TMultilaneGrid::TMultilaneGrid(TParallelID theParallelID) :

ParentClass(ClassInfo, theParallelID)

{
    _Init();
}

/* =================================================================================== */

TMultilaneGrid::~TMultilaneGrid(void)

{
    Dispose();
}

/* =================================================================================== */

bool
TMultilaneGrid::AbsorbVehicles(TMultilaneGrid *Grid)

{
    TVehicle *Vehicle = 0;
    long      CLane, CSite;

    if (GetFirstSite() != 0 OR
        Grid->GetLastSite() != Grid->GetNrOfSites() OR
        GetFirstSite() != 0)
    {
        Error("AbsorbVehicles: cannot absorb from %s", (char *) Grid->GetObjectName());
        return FALSE;
    }
    for (CLane = 0; CLane < Grid->GetNrOfLanes(); CLane++)
        for (CSite = 0; CSite < BoundaryWidth; CSite ++)
            if ((Vehicle = Grid->VAccess(CLane, Grid->LastSite + CSite)))
            {
                
                TRouteplan *Routeplan;

                // migrate the vehicle...
                                
                Vehicle->SetSitePattern(Grid->SAccess(CLane, Grid->LastSite + CSite));
                if (!SetVehicle(CLane, FirstSite + CSite, Vehicle, 
                                Vehicle->GetSitePattern()))
                {
                    Error("AbsorbVehicle: cannot set vehicle");
                    return FALSE;
                }
                Grid->RemoveVehicle(CLane, Grid->LastSite + CSite, FALSE);
                if ((Routeplan = Vehicle->GetRouteplan()))
                    Routeplan->Step();
            }
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::Activate(TDistributedRange &ActiveRange)

{
    if (!AllocateGrids())
    {
        Error("Activate: cannot allocate");
        return FALSE;
    }
    if (Config.NrOfSites < /* 2 * */ GetBndWidth())
    {
        Error("too short! Required length: %ld sites", /* 2 * */ GetBndWidth());
        return FALSE;
    }
    if (Config.NrOfSites > MAX_SITES_PER_GRID)
    {
        Error("too long! Maximum length: %ld sites", MAX_SITES_PER_GRID);
        return FALSE;
    }
    FirstSite = ActiveRange.GetLongBegin(Config.NrOfSites);
    LastSite = ActiveRange.GetLongEnd(Config.NrOfSites);
    Random->SetSeed(1);
    if (FirstSite != 0)
    {
        IncomingBoundaryOffset = FirstSite - BoundaryWidth;
        OutgoingBoundaryOffset = FirstSite;     
    }
    else if (LastSite != Config.NrOfSites)
    {
        IncomingBoundaryOffset = LastSite;
        OutgoingBoundaryOffset = LastSite - BoundaryWidth;
    }
    else
    {
        IncomingBoundaryOffset = INVALID;
        OutgoingBoundaryOffset = INVALID;
    }
    OutgoingMigrationAreaOffset = IncomingBoundaryOffset;
    IncomingMigrationAreaOffset = OutgoingBoundaryOffset;
    DecProbLimit = Random->GetLimit(Config.DecProb);
    ChangeProbLimit[CHANGE_LEFT] = Random->GetLimit(Config.ChangeProb[CHANGE_LEFT]);
    ChangeProbLimit[CHANGE_RIGHT] = Random->GetLimit(Config.ChangeProb[CHANGE_RIGHT]);    
    MarkSites();
    UpdateActiveVehicleRange();
    SpeedLimit = Config.SpeedLimit;
    debug(Print(DM_LIFESPAN, "Activate"));    
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::AllocateGrids(void)

{
    long CLane, CSite;

    if (Config.NrOfSites == INVALID)
    {
        Error("AllocateGrids: invalid NrOfSites in Config");
        return FALSE;
    }
    MoveBoundaryWidth = Config.MoveUpdateSteps * Config.MoveViewRange;
    BoundaryWidth = max(MoveBoundaryWidth, 
                        Config.ChangeUpdateSteps * Config.ChangeViewRange);
    GridArraySize = Config.NrOfSites + 2 * BoundaryWidth;
    if (EncodeBuffer == 0)
        EncodeBuffer = new TSite[GridArraySize];
    for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
        if (Sites[CLane] == 0)
        {
            if ((Sites[CLane] = new TSite[GridArraySize]) == 0 OR
                (Vehicles[CLane] = new TVehicle * [GridArraySize]) == 0)
            {
                Error("AllocateGrids: out of memory");
                return FALSE;
            }
            for (CSite = 0; CSite < GridArraySize; CSite++)
            {
                Sites[CLane][CSite] = 0;
                Vehicles[CLane][CSite] = 0;
            }
        }
    return TRUE;
}

/* =================================================================================== */

void
TMultilaneGrid::AsciiDump(long MaximumWidth)

{
    char Line[256];
    int  Pos;
    long CLane, CSite;
    long InternalSites;

    for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
    {
        Pos = 0; 
        InternalSites = MaximumWidth / 2 - BoundaryWidth;
        for (CSite = 0; CSite < BoundaryWidth; CSite++)
            Line[Pos++] = AsciiSite(SAccess(CLane, FirstSite - BoundaryWidth + CSite));
        Line[Pos++] = (FirstSite == 0)? '|':'B';
        for (CSite = 0; CSite < InternalSites; CSite++)
            Line[Pos++] = AsciiSite(SAccess(CLane, FirstSite + CSite));
        Line[Pos++] = ':';
        for (CSite = 0; CSite < InternalSites; CSite++)
            Line[Pos++] =  AsciiSite(SAccess(CLane, LastSite - InternalSites + CSite));
        Line[Pos++] = (LastSite == Config.NrOfSites)? '|':'B';
        for (CSite = 0; CSite < BoundaryWidth; CSite++)
            Line[Pos++] =  AsciiSite(SAccess(CLane, LastSite + CSite));
        Line[Pos] = 0;
        Print("L%ld %s", CLane, Line);
    }
}

/* =================================================================================== */

bool
TMultilaneGrid::CheckVehicleCount(TTimeStep &TimeStep)

{
    long CLane, CSite;
    long Count;

    Count = 0;
    for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
        for (CSite = FirstSite; CSite < LastSite + BoundaryWidth; CSite++)
            if (VAccess(CLane, CSite) != 0)
                Count++;
    if (Count != Var.OccupiedSites)
    {
        Error("CheckVehicleCount: counted %ld in time-step %s", Count,
              (char *) TimeStep.GetInstanceName());
        return FALSE;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::ComputeOffset(long &Offset, long Width, long InternalBoundarySlack,
                              long ExternalBoundarySlack) const

{
    long   Correction;
    long   StartSlack, EndSlack;

    StartSlack = (FirstSite == 0)? InternalBoundarySlack : ExternalBoundarySlack;
    EndSlack = Width + ((LastSite == Config.NrOfSites)?
                        InternalBoundarySlack : ExternalBoundarySlack);
    if (LastSite - FirstSite < StartSlack + EndSlack)
    {
        Error("CorrectOffset: slack %ld > local length %ld", StartSlack + EndSlack,
              LastSite - FirstSite);
        return FALSE;
    }
    if (Offset > LastSite - EndSlack)
    {
        Correction = Offset - (LastSite - EndSlack);
        Offset -= Correction;
    }
    if (Offset < FirstSite + StartSlack)
    {
        Correction = FirstSite + StartSlack - Offset;
        Offset += Correction;
    }
    if (Offset < FirstSite + StartSlack OR
        Offset > LastSite - EndSlack)
    {
        Error("CorrectOffset: element (width:%ld, offset:%ld) cannot be placed "
              "within [%ld+%ld:%ld-%ld[", Width, Offset, FirstSite, StartSlack,
              LastSite, EndSlack);
        return FALSE;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    if (!ParentClass::Decode(aStream, TransferInfo))
        return FALSE;
    switch (TransferInfo->TransferMode)
    {
    case TM_ConstData:

        if (!Config.Decode(aStream))
            return FALSE;
        break;

    case TM_VarData:
        
        if (!DecodeRange(aStream, &TransferInfo->TransferRange))
        {
            Error("cannot decode range");
            return FALSE;
        }
        break;

    default:

        Warning("invalid transfer mode %d", TransferInfo->TransferMode);
        
    }   
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::DecodeVehicle(TVehicle *& Vehicle, TBaseStream *aStream)

{
    if ((Vehicle = (TVehicle *) ObjectManager->ReceiveObject(aStream, FALSE)) == 0)
    {
        Error("DecodeVehicle: cannot receive");
        return FALSE;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::DecodeRange(TBaseStream *aStream, TDistributedRange *)

{
    long CLane;
    long CSite;
    long Count;
    long i;
    long TransferFirstSite;
    long TransferLastSite;

    if (!AllocateGrids())
    {
        Error("DecodeRange: cannot allocate");  
        return FALSE;
    }
    if (!(aStream->Decode(TransferFirstSite) AND
          aStream->Decode(TransferLastSite)))
        return FALSE;
    if (TransferLastSite >= TransferFirstSite)
    {
        for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
            for (CSite = TransferFirstSite; CSite < TransferLastSite; CSite++)
                InitSite(CLane, CSite);
        for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
        {
            if (!(aStream->Decode(Count) AND
                  aStream->Decode(EncodeBuffer[0], Count)))
                return FALSE;
            for (i = 0; i < Count; i++)
            {
                TVehicle *Vehicle;
                
                CSite = EncodeBuffer[i];
                if (!(DecodeVehicle(Vehicle, aStream)))
                    return FALSE;
                if (Vehicle->Failed())
                    return FALSE;
                SetVehicle(CLane, CSite, Vehicle, Vehicle->GetSitePattern());
            }
        }
    }
    return TRUE;
}

/* =================================================================================== */

void
TMultilaneGrid::Dispose(void)

{
    long CLane;
    long CSite;

    for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
    {
        if (Sites[CLane])
            for (CSite = FirstSite; CSite < LastSite; CSite++)
                RemoveVehicle(CLane, CSite);
        DeleteObjectArray(Sites[CLane]);
        DeleteObjectArray(Vehicles[CLane]);
    }
    DeleteObject(Random);
    DeleteObjectArray(EncodeBuffer);
}

/* =================================================================================== */

bool
TMultilaneGrid::DisposeRange(TDistributedRange *Range)

{
    long DisposeFirstSite;
    long DisposeLastSite;
    long CLane;
    long CSite;

    if (Sites[0])
    {
        DisposeFirstSite = Range->GetLongBegin(Config.NrOfSites);
        DisposeLastSite = Range->GetLongEnd(Config.NrOfSites);  
        for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
            for (CSite = DisposeFirstSite; CSite < DisposeLastSite; CSite++)
                RemoveVehicle(CLane, CSite, TRUE);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::Encode(TBaseStream *aStream, TTransferInfo *TransferInfo) const

{
    if (!ParentClass::Encode(aStream, TransferInfo))
        return FALSE;
    switch (TransferInfo->TransferMode)
    {
    case TM_ConstData:

        if (!Config.Encode(aStream))
            return FALSE;
        break;

    case TM_VarData:

        if (!EncodeRange(aStream, &TransferInfo->TransferRange))
        {
            Error("cannot encode range");
            return FALSE;
        }
        break;

    default:

        Warning("invalid transfer mode %d", TransferInfo->TransferMode);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::EncodeRange(TBaseStream *aStream, TDistributedRange *Range) const

{
    long CSite;
    long CLane;
    long TransferFirstSite;
    long TransferLastSite;
    long Count;

    if (Sites[0])
    {
        TransferFirstSite = Range->GetLongBegin(Config.NrOfSites);
        TransferLastSite = Range->GetLongEnd(Config.NrOfSites);

        // we only transfer what is actually local:

        if (TransferFirstSite < FirstSite)
            TransferFirstSite = FirstSite;
        if (TransferLastSite > LastSite)
            TransferLastSite = LastSite;
    }
    else
    {
        TransferFirstSite = 0;
        TransferLastSite = -1;
    }           
    if (!(aStream->Encode(TransferFirstSite) AND
          aStream->Encode(TransferLastSite)))
        return FALSE;
    
    Print(DM_OBJECT, "EncodeRange [%ld:%ld]", TransferFirstSite, TransferLastSite);

    if (Sites[0])
    {
        for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
        {

            // <<<<< TEMP
            if (LastSite != Config.NrOfSites)
                for (CSite = TransferLastSite; 
                     CSite < TransferLastSite + GetBndWidth(); CSite++)
                    if (VAccess(CLane, CSite))
                        Warning("EncodeRange: after %s: vehicle in migration area %ld:%ld", 
                                (char *) LatestTimeStep.GetInstanceName(), CLane, CSite);
            // >>>>> TEMP
            
            Count = 0;
            for (CSite = TransferFirstSite; CSite < TransferLastSite; CSite++)
                if (SITE_IS_VEHICLE(SAccess(CLane, CSite)))
                {
                    EncodeBuffer[Count++] = CSite;
                    VAccess(CLane, CSite)->SetSitePattern(SAccess(CLane, CSite));
                }
            if (!(aStream->Encode(Count) AND
                  aStream->Encode(EncodeBuffer[0], Count)))
                return FALSE;
            for (CSite = 0; CSite < Count; CSite++)
                if (!EncodeVehicle(VAccess(CLane, EncodeBuffer[CSite]), aStream))
                    return FALSE;
        }
    }
    return TRUE; 
}

/* =================================================================================== */

bool
TMultilaneGrid::EncodeVehicle(TVehicle *Vehicle, TBaseStream *aStream) const

{
    if (!ObjectManager->SendObject(Vehicle, aStream))
    {
        Error("EncodeVehicle: cannot send");
        return FALSE;
    }
    return TRUE;
}

/* =================================================================================== */

void
TMultilaneGrid::ExecuteLaneChanges(void)

{
    long i;
    long ChangeInfo;
    long FromLane;
    long ToLane;

    for (i = 0; i < LaneChangeArray.GetNrOfItems(); i++)
    {
        ChangeInfo = LaneChangeArray[i];
        FromLane  = LC_GET_FROM_LANE(ChangeInfo);
        ToLane = LC_GET_TO_LANE(ChangeInfo);
        if (FromLane < ToLane)
        {
            // left to right

            CombineSites(FromLane, LC_GET_SITE_INDEX(ChangeInfo),
                         ToLane, LC_GET_SITE_INDEX(ChangeInfo));
            if (FromLane >= 2)
                SAccess(FromLane - 2, LC_GET_SITE_INDEX(ChangeInfo)) &= 
                       ~SITE_TMP_NO_CHANGE_LEFT;
            CheckLRVMax(ToLane, LC_GET_SITE_INDEX(ChangeInfo));
        }
        else
        {
            // right to left

            CombineSites(FromLane, LC_GET_SITE_INDEX(ChangeInfo), 
                         ToLane, LC_GET_SITE_INDEX(ChangeInfo));
            CheckRLVMax(ToLane, LC_GET_SITE_INDEX(ChangeInfo));
        }
    }
}

/* =================================================================================== */

bool
TMultilaneGrid::ExchangeBoundaries(TMultilaneGrid *SourceGrid)

{
    long      CLane, CSite;
    int       MaxNrOfLanes;


    if (GetFirstSite() != 0 OR
        SourceGrid->GetLastSite() != SourceGrid->GetNrOfSites() OR
        GetFirstSite() != 0)
    {
        Error("AbsorbVehicles: cannot exchange boundaries with  %s", 
              (char *) SourceGrid->GetObjectName());
        return FALSE;
    }
    if (GetNrOfLanes() > SourceGrid->GetNrOfLanes())

    {
        MaxNrOfLanes = SourceGrid->GetNrOfLanes();

        // destination grid has more lanes than source grid
        
        for (CLane = SourceGrid->GetNrOfLanes(); CLane < GetNrOfLanes(); CLane++)
            for (CSite = 0; CSite < BoundaryWidth; CSite++)
        {
            SAccess(CLane, FirstSite - BoundaryWidth + CSite) = 0;
            VAccess(CLane, FirstSite - BoundaryWidth + CSite) = 0;
        }
    }
    else
    {
        MaxNrOfLanes = GetNrOfLanes();
        if (GetNrOfLanes() < SourceGrid->GetNrOfLanes())
        {
            // destination grid has fewer lanes than source grid

            for (CLane = GetNrOfLanes(); CLane < SourceGrid->GetNrOfLanes(); CLane++)
                SourceGrid->SAccess(CLane, SourceGrid->GetLastSite()) |= SITE_BLOCKED;
        }
    }    
    for (CLane = 0; CLane < MaxNrOfLanes; CLane++)
    {
        for (CSite = 0; CSite < BoundaryWidth; CSite++)
        {
            // copy remote outgoing boundary to local incoming
            
            SAccess(CLane, FirstSite - BoundaryWidth + CSite) = 
                    SourceGrid->SAccess(CLane, SourceGrid->LastSite - BoundaryWidth + CSite);

            /*
            TRouteplan *Routeplan;
            TVehicle   *Vehicle;

            if ((Vehicle = SourceGrid->VAccess(CLane, SourceGrid->LastSite - 
                                               BoundaryWidth + CSite)))
            {
                if ((Routeplan = Vehicle->GetRouteplan()) AND
                    Routeplan->GetCurrentID() != GetParallelID() AND
                    SITE_SourceGrid->SAccess(CLane, SourceGrid->LastSite - BoundaryWidth + CSite)
                    
            }
            */

            // copy local outgoing boundary to remote incoming

            SourceGrid->SAccess(CLane, SourceGrid->LastSite + CSite) = 
                SAccess(CLane, FirstSite + CSite);
            
        }
    }
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneGrid::ExecuteTimeStep(TTimeStep &TimeStep)

{
    int SubStep;

    IfDebug(DM_TIME_STEP, Print("ExecuteTimeStep: %s", (char *) TimeStep.GetInstanceName()));

    SubStep = TimeStep.GetSubStep();
    if (SubStep == 0 AND MoveUpdateRequired())
    {
        if (!CheckVehicleCount(TimeStep))
            return FALSE;
        Move();
        UpdateActiveVehicleRange();
    }
    if (SubStep == 1 AND LaneChangeUpdateRequired())
    {
        LaneChange();
        ExecuteLaneChanges();
    }
    LatestTimeStep = TimeStep;
    return TRUE;
}

// =================================================================================== 

bool
TMultilaneGrid::GetBoundaries(TTimeStep &TimeStep, TDLListPrimitive *Boundaries)

{
    ((TBoundaryList *) Boundaries)->Append(new TMultilaneBoundary(this, TimeStep));
    return TRUE;
}

/* =================================================================================== */

TString
TMultilaneGrid::GetClassName(void) const

{
    return "TMultilaneGrid";
}

/* =================================================================================== */

TString
TMultilaneGrid::GetInstanceName(void) const

{
    char line[256];

    sprintf(line, "(%s) L:%d S:%ld LogR:[%ld,%ld] ActR:[%ld:%ld] Occ:%ld", 
            (char *) ParentClass::GetInstanceName(), Config.NrOfLanes, Config.NrOfSites,
            FirstSite, LastSite - 1, FirstActiveSite, LastActiveSite - 1,
            Var.OccupiedSites);
    return line;
}

// =================================================================================== 
/*
double
TMultilaneGrid::GetLoad(TCPNID)

{
    return (double) (GetNrOfLanes() * GetNrOfLocalSites());
}
*/
// =================================================================================== 

void
TMultilaneGrid::_Init(void)

{
    int CLane;

    Random = new TRandom();
    EncodeBuffer = 0;
    FirstSite = LastSite = INVALID;
    FirstActiveSite = LastActiveSite = INVALID;
    for (CLane = 0; CLane < MAX_LANES; CLane++)
    {
        Sites[CLane] = 0;
        Vehicles[CLane] = 0;
        BoundaryBufferOffsets[CLane] = INVALID;
    }
    ResetBoundaryBuffer();
}

/* =================================================================================== */

bool
TMultilaneGrid::LaneChangeUpdateRequired(void) const

{
    return (Var.OccupiedSites > 0 OR !SkipWhenEmpty()) AND 
        Config.NrOfLanes > 1 AND (Config.Flags & CA_PASSING_PROHIBITION) == 0;
}

/* =================================================================================== */

void
TMultilaneGrid::LoadBoundaryBuffer(void)

{
    long CLane;

    for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
        if (BoundaryBufferOffsets[CLane] != INVALID)
            if (SITE_IS_VEHICLE(IncomingBoundary(CLane, BoundaryBufferOffsets[CLane])))
            {
                if (BoundaryBufferSites[CLane] != 
                    IncomingBoundary(CLane, BoundaryBufferOffsets[CLane]))
                    Warning("LoadBoundaryBuffer: %s pattern mismatch in site %ld:%ld "
                            "buffer:%ld boundary:%ld", 
                            (char *) LatestTimeStep.GetInstanceName(),
                            CLane, BoundaryBufferOffsets[CLane],
                            BoundaryBufferSites[CLane],
                            IncomingBoundary(CLane, BoundaryBufferOffsets[CLane]));
            }
            else
                IncomingBoundary(CLane, BoundaryBufferOffsets[CLane]) = 
                    BoundaryBufferSites[CLane];
}

// =================================================================================== 

void
TMultilaneGrid::MarkSites(void)

{
    long CLane;
    long CSite;

    // Print("InitSites with Flags %ld", Config.Flags);

    if (Config.Flags & CA_BLOCK_SOURCE)
        for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
            for (CSite = 0; CSite < BoundaryWidth; CSite++)
                FromBoundary(CLane, CSite) |= SITE_BLOCKED;
    if (Config.Flags & CA_BLOCK_DEST)
        for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
            for (CSite = -CLane * Config.MaxV; CSite < BoundaryWidth; CSite++)
                ToBoundary(CLane, CSite) |= SITE_BLOCKED;
}

/* =================================================================================== */

bool
TMultilaneGrid::MoveUpdateRequired(void) const

{
    return Var.OccupiedSites > 0 OR !SkipWhenEmpty();
}

// =================================================================================== 

bool
TMultilaneGrid::ReadEvent(TStream *Stream)

{
    char EventName[128];
    long NewSpeedLimit;

    if (!Stream->Read(EventName))
    {
        Error("ReadEvent: cannot read event name");
        return FALSE;
    }
    if (strcmp(EventName, "SetSpeedLimit") == 0)
    {
        if (!Stream->Decode(NewSpeedLimit))
        {
            Error("ReadEvent: cannot decode new SpeedLimit");
            return FALSE;
        }
        Print("changing speed limit from %ld to %ld", SpeedLimit, NewSpeedLimit);
        SpeedLimit = NewSpeedLimit;
        return TRUE;
    }
    if (strcmp(EventName, "RestoreSpeedLimit") == 0)
    {
        SpeedLimit = Config.SpeedLimit;
        Print("restoring previous speed limit %ld", SpeedLimit);
        return TRUE;
    }
    Warning("ReadEvent: unknown event name '%s'", EventName);   
    return TRUE;
}

/* =================================================================================== */

void
TMultilaneGrid::ResetBoundaryBuffer(void)

{
    long CLane;

    for (CLane = 0; CLane < MAX_LANES; CLane++)
        BoundaryBufferOffsets[CLane] = INVALID;
}

/* =================================================================================== */

void
TMultilaneGrid::UpdateActiveVehicleRange(void)

{
    /*
    long CLane, CSite, Found;

    if (Var.OccupiedSites > 3)
    {
    */
        FirstActiveSite = FirstSite;
        LastActiveSite = LastSite;
        /*
    }
    else
    {
        // if (!SkipWhenEmpty()) 

        // move the lower limit forward if possible
        
        Found = FALSE;
        CSite = FirstActiveSite;
        while (CSite < LastActiveSite AND !Found)
        {
            CLane = 0;
            while (CLane < Config.NrOfLanes AND !Found)
            {       
                if (SITE_IS_VEHICLE(SAccess(CLane, CSite)))
                {
                    Found = TRUE;
                    CSite--;
                }
                else
                    CLane++;
            }
            CSite++;
        }
        FirstActiveSite = CSite;
        
        // move the upper limit backward if possible
        
        Found = FALSE;
        CSite = LastActiveSite + Config.MaxV;
        if (CSite > LastSite - 1)
            CSite = LastSite - 1;
        while (CSite >= FirstActiveSite AND !Found)
        {
            CLane = 0;
            while (CLane < Config.NrOfLanes AND !Found)
            {       
                if (SITE_IS_VEHICLE(SAccess(CLane, CSite)))
                {
                    Found = TRUE;
                    CSite++;
                }
                else
                    CLane++;
            }
            CSite--;
        }
        LastActiveSite = CSite + 1;
    }
    if (LastActiveSite < FirstActiveSite)
        Warning("LastActiveSite < FirstActiveSite");
        */
}

/* =================================================================================== */

bool
TMultilaneGrid::UpdateRequired(void) const

{
    return MoveUpdateRequired() OR LaneChangeUpdateRequired();
}

/* =================================================================================== */
#undef ParentClass
