/*      file     : TrafficData/TMultilaneGrid.h
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TMultilaneGrid.h,v 1.11 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#ifndef TMULTILANEGRID_H
#define TMULTILANEGRID_H

// ***INCLUDES************************************************************************

#include <CAConfig.h>

#include <TParallelObject.h>

#include <TVehicle.h>  // This should not be here!
#include <TTimeStep.h> // Neither should this!

#include "TMultilaneGridConfig.h"

// ***FORWARDS************************************************************************

class TBaseStream;
class TClassInfo;
class TDistributedRange;
class TMultilaneStat;
// class TTimeStep;
class TTransferInfo;
class TRandom;

// ***DEFINES*************************************************************************

#define LOOKUP_TABLE_SIZE (8*8*8*2)

// defines to extract site index and lane numbers from lane changing info

#  define _SAccess(Lane, Site) Sites[Lane][BoundaryWidth + (Site)]
#  define _VAccess(Lane, Site) Vehicles[Lane][BoundaryWidth + (Site)]

// ***CLASSES*************************************************************************

class TMultilaneGridVarData

{
public:

    TMultilaneGridVarData(void);

    virtual bool
        Decode(TBaseStream *aStream);
    virtual bool
        Encode(TBaseStream *aStream) const;

    long          OccupiedSites;
    int           Seed;    
};

// ===================================================================================

class TMultilaneGrid : public TParallelObject

{
    friend class TMultilaneDump;
    friend class TMultilaneStat;
    friend class TMultilaneBoundary;
    friend class TGridStat;
    
 public:

    TMultilaneGrid(TParallelID theParallelID);

    virtual ~TMultilaneGrid(void);
    
    bool
    AbsorbVehicles(TMultilaneGrid *Grid);

    bool
    Activate(TDistributedRange &ActiveRange);

    void
    AsciiDump(long MaximumWidth);

    inline void
    AssignConfig(const TMultilaneGridConfig *theConfig);

    bool
    ComputeOffset(long &Offset, long Width,
                  long InternalBoundaryWidth, long ExternalBoundaryWidth) const;

    virtual bool
    Decode(TBaseStream *, TTransferInfo *TransferInfo);

    virtual bool
    Encode(TBaseStream *, TTransferInfo *TransferInfo) const;

    bool
    ExchangeBoundaries(TMultilaneGrid *Grid);

    bool
    ExecuteTimeStep(TTimeStep &TimeStep);

    inline TSite &
    FromBoundary(long Lane, long Offset) const;

    inline int
    GetBndWidth(void) const;

    virtual TString
    GetInstanceName(void) const;

    inline long
    GetFirstSite(void) const;

    inline long
    GetLastSite(void) const;

    inline int
    GetNrOfLanes(void) const;

    inline long
    GetNrOfActiveSites(void) const;

    inline long
    GetNrOfLocalSites(void) const;

    inline long
    GetNrOfOccupiedSites(void) const;

    inline long
    GetNrOfSites(void) const;

    inline bool
    HasMigratingVehicles(const TTimeStep &TimeStep) const;

    inline TSite &
    IncomingBoundary(long Lane, long Offset) const;

    inline TSite &
    IncomingMigrationArea(long Lane, long Offset) const;

    inline TVehicle * &
    IncomingVehicleMigrationArea(long Lane, long Offset) const;

    void
    LoadBoundaryBuffer(void);

    inline TSite &
    OutgoingBoundary(long Lane, long Offset) const;

    inline TSite &
    OutgoingMigrationArea(long Lane, long Offset) const;

    inline TVehicle * &
    OutgoingVehicleMigrationArea(long Lane, long Offset) const;

    inline void
    RemoveVehicle(long Lane, long Site, bool DisposeVehicle = TRUE);
    
    inline void
    RemoveVehicleFromOutgoingMigrationArea(long Lane, long Site);

    void
    ResetBoundaryBuffer(void);

    inline TSite &
    SAccess(long Lane, long Position) const;

    inline void
    SetBoundaryBuffer(long CLane, long Offset, TSite Site);

    inline bool
    SetMigratedVehicle(long Lane, long Offset, TVehicle *Vehicle, TSite SitePattern);

    inline bool
    SetVehicle(long Lane, long Offset, TVehicle *Vehicle, TSite SitePattern);

    inline TSite &
    ToBoundary(long Lane, long Offset) const;

    inline TVehicle * &
    VAccess(long Lane, long Position) const;

    inline bool
    VehicleRangeIsActive(long Offset, long Width) const;


    static TClassInfo *ClassInfo;

private:


    bool
    AllocateGrids(void);

    inline char
    AsciiSite(TSite Site) const;

    inline void
    CheckLRVMax(long Lane, long Site) const;

    inline void
    CheckRLVMax(long Lane, long Site) const;

    inline void
    CheckActiveVehicleRangeAfterDeletion(void);

    inline void
    CheckActiveVehicleRangeAfterInsertion(long Offset);

    bool
    CheckVehicleCount(TTimeStep &TimeStep);

    bool
    DecodeRange(TBaseStream *aStream, TDistributedRange *Range);

    bool
    DecodeVehicle(TVehicle *&Vehicle, TBaseStream *aStream);

    void
    Dispose(void);

    virtual bool
    DisposeRange(TDistributedRange *Range);

    bool
    EncodeRange(TBaseStream *aStream, TDistributedRange *Range) const;

    bool
    EncodeVehicle(TVehicle *Vehicle, TBaseStream *aStream) const;

    void
    ExecuteLaneChanges(void);

    virtual bool
    GetBoundaries(TTimeStep &, TDLListPrimitive *Boundaries);

    inline int
    GetChangeViewRange(void) const;

    virtual TString
    GetClassName(void) const;
    
    //virtual double
    //GetLoad(TCPNID LocalCPNID);

    inline int
    GetMoveBndWidth(void) const;

    inline int
    GetMoveViewRange(void) const;

    inline int
    GetReducedGap(long Lane) const;

    inline bool
    HasReducedGap(long Lane) const;

    void
    _Init(void);

    void
    MarkSites(void);

    inline bool
    IsAsymetric() const;

    void
    LaneChange(void);

    bool
    LaneChangeUpdateRequired(void) const;

    void
    Move(void);

    bool
    MoveUpdateRequired(void) const;

    virtual bool
    ReadEvent(TStream *Stream);

    void
    ResetStatistics(void);

    inline bool
    SkipWhenEmpty(void) const;

    bool
    UpdateRequired(void) const;

    // Methods for boundary access

    inline TSite &
        LeftBoundary(long Lane, long Site) const;
    inline TSite &
        LeftBoundaryArea(long Lane, long Site) const;
    inline TSite &
        OutboundLeftBoundary(long Lane, long Site) const;
    inline TSite &
        OutboundRightBoundary(long Lane, long Site) const;
    inline TSite &
        RightBoundary(long Lane, long Site) const;

    // Methods for CA update

    inline void
        CombineSites(long Lane, long Site,long ToLane, long ToSite) const;
    inline long
        CorLane(long Lane) const;
    inline void
        InitSite(long Lane, long Site) const;
    inline void
        MoveEmptySite(long Lane, long Site, long ToLane, long ToSite) const;
    inline void
        MoveSite(long Lane, long Site, long ToLane, long ToSite) const;
    inline bool
    RandomDeceleration(TSite Site) const;
    inline TSite &
        SRawAccess(long Lane, long Position) const;
    inline void
        UpdateSite(long Lane, long From, long NewV) const;
    void
        UpdateActiveVehicleRange(void);    

    TTimeStep     LatestTimeStep; // !!!!! TEMP

    TMultilaneGridConfig  Config;
    TMultilaneGridVarData Var;

    long          SpeedLimit;

    TRandom      *Random;
    TSite        *Sites[MAX_LANES];
    TSite        *EncodeBuffer;
    TVehicle    **Vehicles[MAX_LANES];
    long          BoundaryBufferOffsets[MAX_LANES];
    TSite         BoundaryBufferSites[MAX_LANES];

    long          GridArraySize;
    long          FirstSite;
    long          LastSite;
    long          FirstActiveSite;
    long          LastActiveSite;
    long          IncomingBoundaryOffset;
    long          OutgoingBoundaryOffset;
    long          IncomingMigrationAreaOffset;
    long          OutgoingMigrationAreaOffset;

    TArray<long>  LaneChangeArray;
    long          DecProbLimit;
    long          ChangeProbLimit[2];
    bool          BoundaryExternal[2];
    int           BoundaryWidth;
    int           MoveBoundaryWidth;

    long          SimpleStatSampleCount;
    long          SimpleStatVCount;
    long          SimpleStatVSum;
    long          SimpleStatVSumSquare;
};

// ***INLINE METHODS******************************************************************

#ifndef NO_INLINE
#  include "TMultilaneGrid.i.C"
#endif

// ===================================================================================
// Local Variables:
// mode:c++
// End:
#endif


