/*      file     : TrafficData/TShortestPath.C
        project  : PAMINA III
        authors  : Marcus Rickert
        RCS      : $Id: TShortestPath.C,v 1.5 2005/03/25 12:28:52 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdlib.h>

#include <TStream.h>
#include <TTimeStep.h>
#include <TNode.h>
#include <TEdge.h>
#include <TIncidenceInfo.h>
#include <TConfiguration.h>
#include <TParallel.h>

#include "TrafficDataConfig.h"
#include "TRouteplan.h"
#include "TVehicle.h"
#include "TMultilaneGrid.h"
#include "TGridStatLookup.h"
#include "TShortestPath.h"

// ***DEFINES************************************************************************* 

#define WriteIfActive(Stream, String) if (Stream) Stream->Write(String)

#define SP_DEFAULT_SOURCETIMESLACK  0L
#define SP_DEFAULT_MANHATTENLIMIT   ((double) INVALID)

// ***FUNCTIONS*********************************************************************** 

int
CompareAlgoInfos(TAlgoInfo *Info1, TAlgoInfo *Info2)

{
    if (Info1->GetParallelID() < Info2->GetParallelID())
        return -1;
    else
        if (Info1->GetParallelID() > Info2->GetParallelID())
            return 1;
        else
            return 0;
}

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TShortestPath::TShortestPath(long theStartTimeStep, TClassInfo *theLookupClassInfo)

{
    StartTimeStep = theStartTimeStep;
    LookupClassInfo = theLookupClassInfo;
    AlgoInfos = 0;
    AlgoEdges = 0;
    Incidences = 0;
    LookupClassInfo->RegisterActionHandler(this);
    Configuration->SetDefault(TConfigEntry(SP_GROUP, SP_NAME_SOURCETIMESLACK,
                                           SP_DEFAULT_SOURCETIMESLACK));    
    Configuration->SetDefault(TConfigEntry(SP_GROUP, SP_NAME_MANHATTENLIMIT,
                                           SP_DEFAULT_MANHATTENLIMIT));
    SourceTimeSlack = Configuration->GetLong(SP_GROUP, SP_NAME_SOURCETIMESLACK);
    ManhattenLimit = Configuration->GetDouble(SP_GROUP, SP_NAME_MANHATTENLIMIT);
}

// =================================================================================== 

TShortestPath::~TShortestPath(void)

{
    Deallocate();
}

// =================================================================================== 

bool
TShortestPath::Allocate(void)

{
    TNodeEnum CNode;
    TEdgeEnum CEdge;
    int       CInfo;
    long      CIndex;
    TInfoID   LookupInfoID;
    TInfoID   IncInfoID;

    if (AlgoInfos)
        return TRUE;
    if (!DetermineLocalNetwork())
    {
        Error("Allocate: cannot determine local network");
        goto OnError;
    }
    if (Nodes.GetNrOfElements() == 0 OR Edges.GetNrOfElements() == 0)
    {
        Error("AllocateArrays: cannot allocate arrays for %ld nodes and %ld edges",
              Nodes.GetNrOfElements(), Edges.GetNrOfElements());
        goto OnError;
    }
    
    AlgoInfos = new TAlgoInfo[Nodes.GetNrOfElements()];
    AlgoEdges = new TAlgoEdge[Edges.GetNrOfElements()];
    Incidences = new long[Edges.GetNrOfElements()];

    CNode.WalksForwardThrough(Nodes);
    CIndex = 0;
    while (CNode())
    {
        AlgoInfos[CIndex++].Load((TNode *) CNode);
        CNode++;
    }
    qsort((void *) AlgoInfos, Nodes.GetNrOfElements(), sizeof(TAlgoInfo),
          (int (*) (const void *, const void *)) CompareAlgoInfos);
    LookupInfoID = LookupClassInfo->GetInfoID();
    CEdge.WalksForwardThrough(Edges);
    CIndex = 0;
    while (CEdge())
    {
        TNode     *SourceNode, *DestNode;
        TAlgoInfo *SourceAlgoNode, *DestAlgoNode;

        if ((SourceNode = CEdge->GetFromNode()) == 0)
        {
            Error("Allocate: cannot retrieve FromNode");
            goto OnError;
        }
        if ((SourceAlgoNode = GetAlgoInfo(SourceNode, TRUE)) == 0)
        {
            Error("Allocate: cannot retrieve source algo info");
            goto OnError;
        }
        if ((DestNode = CEdge->GetToNode()) == 0)
        {
            Error("Allocate: cannot retrieve ToNode");
            goto OnError;
        }
        if ((DestAlgoNode = GetAlgoInfo(DestNode, TRUE)) == 0)
        {
            Error("Allocate: cannot retrieve dest algo info");
            goto OnError;
        }
        AlgoEdges[CIndex++].Load(CEdge, LookupInfoID, SourceAlgoNode, DestAlgoNode);
        CEdge++;
    }
    qsort((void *) AlgoEdges, Edges.GetNrOfElements(), sizeof(TAlgoEdge),
          (int (*) (const void *, const void *)) CompareAlgoEdges);
    IncInfoID = TIncidenceInfo::ClassInfo->GetInfoID();
    CIndex = 0;
    CInfo = 0; 
    while (CInfo < Nodes.GetNrOfElements() AND 
           CIndex < Edges.GetNrOfElements())
    {
        TIncidenceInfo *Inc;
        int             COutEdge;
        long            FirstIncidentEdgeIndex;

        if ((Inc = (TIncidenceInfo *) 
             AlgoInfos[CInfo].GetNode()->GetInfo(IncInfoID, TRUE, TRUE)) == 0)
        {
            Error("Allocate: cannot load incidence info for %s",
                  (char *) AlgoInfos[CInfo].GetNode()->GetObjectName());
            goto OnError;
        }
        FirstIncidentEdgeIndex = CIndex;
        COutEdge = 0;
        while (COutEdge < Inc->GetNrOfOutgoingEdges() AND 
               CIndex < Edges.GetNrOfElements())
        {
            int    AlgoEdgeIndex = 0;
            TEdge *OutgoingEdge;
            
            if ((OutgoingEdge = Inc->GetOutgoingEdge(COutEdge, FALSE)) AND
                (AlgoEdgeIndex = GetAlgoEdgeIndex(OutgoingEdge)) != INVALID)
                Incidences[CIndex++] = AlgoEdgeIndex;
            COutEdge++;
        }
        AlgoInfos[CInfo].SetIncidences(FirstIncidentEdgeIndex, CIndex - FirstIncidentEdgeIndex);
        CInfo++;
    }
    if (CIndex == Edges.GetNrOfElements() AND CInfo < Nodes.GetNrOfElements())
    {
        Error("Allocate: cannot store all incidences");
        goto OnError;
    }
    if (CIndex < Edges.GetNrOfElements())
        Warning("Allocate: %ld outgoing incidences could not be retrieved", 
                Edges.GetNrOfElements() - CIndex);
    PreviousAccessInfo.Node = 0;
    return TRUE;

OnError:

    LatestError = SP_CannotAllocate;
    Deallocate();
    return FALSE;
}

// =================================================================================== 

bool
TShortestPath::BuildTree(TNetAccessInfo *AccessInfos, int NrOfAccessInfos)

{
    TAlgoInfo          *ScanInfo;
    TAlgoInfo          *OtherInfo;
    TNodeList           ScanList;
    int                 CEdgeIndex;
    TAlgoEdge          *CEdge;
    THeap< TAlgoInfo >  MinHeap;
    DistType            NewDist;
    DistType            DistToScanNode;
    TTimeStep           CurrentEdgeTime;
    long                NodesLabeled;
    int                 CAccessInfo;
    TLocation           SourceLocation;

    LatestError = SP_OK;
    if (!Allocate())
    {
        Error("BuildTree: cannot allocate");
        return FALSE;
    }
    SourceTimeOffset = 0;
    if (!ProcessNetAccessInfos(AccessInfos, NrOfAccessInfos, TRUE))
        return FALSE;
    SourceLocation = AccessInfos[0].AlgoInfo->GetLocation();
    if (NrOfAccessInfos == 1)
        if (AccessInfos[0].Node == PreviousAccessInfo.Node AND
            abs(AccessInfos[0].AlgoInfoTime - PreviousAccessInfo.AlgoInfoTime) <= SourceTimeSlack)
        {
            SourceTimeOffset = AccessInfos[0].AlgoInfoTime - PreviousAccessInfo.AlgoInfoTime;
            //      Print("re-using old tree with offset %ld", SourceTimeOffset);
            return TRUE;
        }
    ResetAlgoInfos();
    for (CAccessInfo = 0; CAccessInfo < NrOfAccessInfos; CAccessInfo++)
    {
        AccessInfos[CAccessInfo].AlgoInfo->SetDist(AccessInfos[CAccessInfo].AlgoInfoTime);
        MinHeap.Add(AccessInfos[CAccessInfo].AlgoInfo);

        //      Print("set source with start distance %ld", AccessInfos[CAccessInfo].AlgoInfoTime);

    }
    NodesLabeled = 0;
    while (MinHeap.GetNrOfElements())
    {
        ScanInfo = MinHeap.GetTop();
        NodesLabeled++;
        DistToScanNode = ScanInfo->GetDist();
        CurrentEdgeTime = (long) DistToScanNode;

        //      Print("distance to scan node %f", DistToScanNode);

        for (CEdgeIndex = 0; CEdgeIndex < ScanInfo->GetNrOfIncidentEdges(); CEdgeIndex++)
        {

            CEdge = AlgoEdges + Incidences[ScanInfo->GetFirstIncidentEdgeIndex() + CEdgeIndex];
            OtherInfo = CEdge->GetDestAlgoInfo();
            if (OtherInfo->IsInHeap())
            {           
                double EdgeTravelTime;
                
                if ((EdgeTravelTime = 
                     CEdge->Lookup->GetAverageTravelTime(CurrentEdgeTime)) != INVALID)
                {
                    NewDist = DistToScanNode + EdgeTravelTime;
                    if (OtherInfo->Improved(NewDist, CEdge))
                        MinHeap.Update(OtherInfo);
                }

                //              Print("EdgeTraveTime %f", EdgeTravelTime);

            }
            else
                if (OtherInfo->GetLocation().ManhattenDistanceTo(SourceLocation) < ManhattenLimit AND
                    OtherInfo->GetDist() == INVALID_DIST)
                {
                    // Print("weight: %f", CEdge->GetWeight(ScanNode));
                    
                    double EdgeTravelTime;
                    
                    if ((EdgeTravelTime = 
                         CEdge->Lookup->GetAverageTravelTime(CurrentEdgeTime)) != INVALID)
                    {
                        NewDist = DistToScanNode + EdgeTravelTime;
                        OtherInfo->SetDist(NewDist);
                        OtherInfo->SetPred(CEdge);
                        MinHeap.Add(OtherInfo);
                    }

                    //              Print("EdgeTravelTime %f", EdgeTravelTime);

                }
        }
    }
    //    Print("%ld nodes labeled", NodesLabeled);
    PreviousAccessInfo = AccessInfos[0];
    return TRUE;
}

// ===================================================================================

void
TShortestPath::Deallocate(void)

{
    DeleteObjectArray(AlgoInfos);
    DeleteObjectArray(AlgoEdges);
    DeleteObjectArray(Incidences);
}

// ===================================================================================

bool
TShortestPath::DetermineLocalNetwork(void)

{
    TEnumerator< TGridStatLookup > CLookup(LookupClassInfo->GetTable());

    Edges.DeleteAll();
    CLookup.Enumerate();
    while (CLookup())
    {
        TEdge *Edge;
        TNode *Node;

        if ((Edge = TEdge::GetPtr(CLookup->GetParallelID())) == 0)
        {
            Error("DetermineLocalNetwork: cannot retrieve edge for %s", 
                  (char *) CLookup->GetObjectName());
            return FALSE;
        }
        Edges.Append(Edge);
        if ((Node = Edge->GetFromNode()) == 0)
        {
            Error("DetermineLocalNetwork: cannot retrieve FromNode for %s",
                  (char *) Edge->GetObjectName());
            return FALSE;
        }
        Nodes.Append(Node);
        if ((Node = Edge->GetToNode()) == 0)
        {
            Error("DetermineLocalNetwork: cannot retrieve ToNode for %s",
                  (char *) Edge->GetObjectName());
            return FALSE;
        }
        Nodes.Append(Node);
        CLookup++;
    }
    Print("found %ld nodes and %ld edges in local grid statistics network",
          Nodes.GetNrOfElements(), Edges.GetNrOfElements());
    return TRUE;
}

// ===================================================================================

bool
TShortestPath::EstimateRemainingTravelTime(long CurrentTimeStep, TParallelID CurrentEdgeID,
                                           double RelativePosition, TRouteplan *Route,
                                           int &LastValidStep, long &EstimatedArrivalTimeStep)

{
    int        CStep;
    double     RemainingTravelTime;
    TTimeStep  LocalVehicleTime;
    double     EdgeTravelTime;
    long       Delay;
    TEdge     *CurrentEdge;
    TInfoID    LocationInfoID;
    TLocation  SourceLocation;

    LastValidStep = Route->GetCurrentStep() - 1;
    LocalVehicleTime = CurrentTimeStep;
    if ((CurrentEdge = TEdge::GetPtr(CurrentEdgeID)) == 0)
    {
        Error("EstimateRTT: cannot retrieve current edge %ld", CurrentEdgeID);
        return FALSE;
    }
    if ((EdgeTravelTime = GetEdgeTravelTime(CurrentEdge, LocalVehicleTime)) == INVALID)
    {
        Error("EstimateRTT: cannot retrieve TT for current edge %ld", CurrentEdgeID);
        return FALSE;
    }
    SourceLocation = CurrentEdge->GetFromNode()->Loc(TLocationInfo::ClassInfo->GetInfoID());
    EdgeTravelTime = (1.0 - RelativePosition) * EdgeTravelTime;
    RemainingTravelTime = EdgeTravelTime;
    LocalVehicleTime += (long) EdgeTravelTime;
    LocationInfoID = TLocationInfo::ClassInfo->GetInfoID();
    for (CStep = Route->GetCurrentStep(); CStep < Route->GetNrOfSteps(); CStep++)
    {
        TEdge  *Edge;
        TNode  *ToNode;
        double  ManhattenDist;
        
        if ((Edge = TEdge::GetPtr(Route->GetStep(CStep).ID)) == 0)
        {
            Warning("EstimateRTT: cannot retrieve edge %ld", Route->GetStep(CStep).ID);
            RemainingTravelTime += Route->GetArrivalTimeStep() - Route->GetStep(CStep).Estimate;
            break;
        }
        ToNode = Edge->GetFromNode();
        ManhattenDist = ToNode->Loc(LocationInfoID).ManhattenDistanceTo(SourceLocation);
        if (ManhattenDist > ManhattenLimit)
        {
            //      Warning("EstimateRTT: Manhatten distance exceeded (%f > %f)", 
            //              LengthInKilometer(ManhattenDist), LengthInKilometer(ManhattenLimit));
            RemainingTravelTime += Route->GetArrivalTimeStep() - Route->GetStep(CStep).Estimate;
            break;
        }
        if ((EdgeTravelTime = GetEdgeTravelTime(Edge, LocalVehicleTime)) == INVALID)
        {
            RemainingTravelTime += Route->GetArrivalTimeStep() - Route->GetStep(CStep).Estimate;
            Print("EstimateRTT: using plan estimate for remaining part of route %s",
                  (char *) Route->GetInstanceName());
            break;
        }

        /*
        if (EdgeTravelTime > 300)
            Print("edge:%ld tt:%.1f", Route->GetStep(CStep).ID, EdgeTravelTime);
            */

        if (CStep == Route->GetNrOfSteps() - 1 AND Route->DestinationIsAccessory())
            EdgeTravelTime /= 2.0;
        RemainingTravelTime += EdgeTravelTime;
        LocalVehicleTime += (long) EdgeTravelTime;
        LastValidStep = CStep;
    }
    EstimatedArrivalTimeStep = CurrentTimeStep + (long) RemainingTravelTime;
    Delay = EstimatedArrivalTimeStep - Route->GetEstimatedDeletionTimeStep();
    
    //    Print("current:%ld remaining:%ld insertion:%ld estimate:%ld",
    //    (long) CurrentTime, (long) RemainingTravelTime,
    //    Route->GetInsertionTimeStep(), Route->GetEstimatedTravelTime());

    Route->SetEstimatedRelativeDelay(1.0 * Delay / Route->GetEstimatedTravelTime());
    return TRUE;
}

// =================================================================================== 

long
TShortestPath::GetAlgoEdgeIndex(TEdge *Edge, bool IssueWarning)

{
    TAlgoEdge *Result;

    SearchEdge.Load(Edge);
    
    if ((Result = (TAlgoEdge *) 
         bsearch((void *) &SearchEdge, AlgoEdges, Edges.GetNrOfElements(),
                 sizeof(TAlgoEdge),  (int (*) (const void *, const void *)) 
                 CompareAlgoEdges)) == 0)
    {
        if (IssueWarning)
            Warning("GetAlgoEdgeIndex: cannot find %s", (char *) Edge->GetObjectName());
        return INVALID;
    }
    return Result - AlgoEdges;
}

// =================================================================================== 

TAlgoInfo *
TShortestPath::GetAlgoInfo(TNode *Node, bool IssueWarning)

{
    TAlgoInfo *Result;

    SearchNode.Load(Node);    
    if ((Result = (TAlgoInfo *) 
         bsearch((void *) &SearchNode, AlgoInfos, Nodes.GetNrOfElements(),
                 sizeof(TAlgoInfo),  (int (*) (const void *, const void *)) 
                 CompareAlgoInfos)) == 0)
    {
        if (IssueWarning)
            Warning("GetAlgoInfo: cannot find %s", (char *) Node->GetObjectName());
        return 0;
    }
    return Result;
}

// =================================================================================== 

long
TShortestPath::GetAlgoInfoIndex(TParallelID SearchID, bool IssueWarning)

{
    TAlgoInfo *Result;

    SearchNode.Load(SearchID);
    if ((Result = (TAlgoInfo *) 
         bsearch((void *) &SearchNode, AlgoInfos, Nodes.GetNrOfElements(),
                 sizeof(TAlgoInfo),  (int (*) (const void *, const void *)) 
                 CompareAlgoInfos)) == 0)
    {
        if (IssueWarning)
            Warning("GetAlgoInfoIndex: cannot find node %ld", SearchID);
        return INVALID;
    }
    return Result - AlgoInfos;
}

// =================================================================================== 

TString
TShortestPath::GetClassName(void) const

{
    return "TShortestPath";
}

// =================================================================================== 

double
TShortestPath::GetEdgeTravelTime(TEdge *Edge, long TimeStep)

{
    TGridStatLookup *Lookup;
    TInfoID          GridStatInfoID;

    GridStatInfoID = LookupClassInfo->GetInfoID();
    if ((Lookup = (TGridStatLookup *) Edge->GetInfo(GridStatInfoID)) == 0)
        return (double) INVALID;
    return Lookup->GetAverageTravelTime(TimeStep);
}

// =================================================================================== 

TShortestPathError
TShortestPath::GetError(void) const

{
    return LatestError;
}

// =================================================================================== 

bool
TShortestPath::GetPath(TNetAccessInfo *AccessInfos, int NrOfAccessInfos,
                       TRouteStep *RouteSteps, int MaxRouteSteps, 
                       int &NrOfRouteSteps, TNetAccessInfo *&DestAccessInfo)

{
    TAlgoInfo *CurrentInfo;
    int        CInfo;
    int        CStep;

    LatestError = SP_OK;
    if (!ProcessNetAccessInfos(AccessInfos, NrOfAccessInfos, FALSE))
        return FALSE;

    // first, find the best possible access to the destination

    DestAccessInfo = 0;
    for (CInfo = 0; CInfo < NrOfAccessInfos; CInfo++)
        if (DestAccessInfo == 0 OR
            AccessInfos[CInfo].Time < DestAccessInfo->Time)
            DestAccessInfo = AccessInfos + CInfo;

    if (DestAccessInfo == 0 OR
        DestAccessInfo->AlgoInfo->GetDist() == INVALID_DIST)
    {
        LatestError = SP_NoPath;
        return FALSE;
    }
    
    // second, count the number of edges

    NrOfRouteSteps = 0;
    CurrentInfo = DestAccessInfo->AlgoInfo;
    while (CurrentInfo AND NrOfRouteSteps <= MaxRouteSteps)
    {
        TAlgoEdge *Pred;

        Pred = CurrentInfo->GetPred();
        if (Pred)
        {
            CurrentInfo = Pred->GetOriginAlgoInfo();
            NrOfRouteSteps++;
        }
        else
            CurrentInfo = 0;    
    }
    if (NrOfRouteSteps > MaxRouteSteps)
    {
        LatestError = SP_PathTooLong;
        return FALSE;
    }

    // third, invert the path and store it

    CurrentInfo = DestAccessInfo->AlgoInfo;
    CStep = NrOfRouteSteps - 1;
    while (CurrentInfo)
    {
        TAlgoEdge *Pred;

        Pred = CurrentInfo->GetPred();
        if (Pred)
        {
            CurrentInfo = Pred->GetOriginAlgoInfo();
            RouteSteps[CStep].ID = Pred->Edge->GetParallelID();
            RouteSteps[CStep].Estimate = (long) CurrentInfo->GetDist();
            CStep--;
        }
        else
            CurrentInfo = 0;    
    }
    return TRUE;
}

// =================================================================================== 

bool
TShortestPath::HandleAction(TParallelObject *theParallelObject, TObjectAction theAction)

{
    if (theParallelObject->GetClassInfo() == LookupClassInfo)
        return TRUE;
    return TObjectHandler::HandleAction(theParallelObject, theAction);
}

// =================================================================================== 

bool
TShortestPath::HandleAnonymousAction(TClassInfo *theClassInfo, TObjectAction theAction)

{
    if (theClassInfo == LookupClassInfo)
    {
        switch(theAction)
        {
        case ActionAdd:
            
            Deallocate();
            break;

        default:

            break;
        }
        return TRUE;
    }
    return TObjectHandler::HandleAnonymousAction(theClassInfo, theAction);
}

// =================================================================================== 

bool
TShortestPath::IsActive(void) const

{
    return (AlgoInfos != 0);
}

// =================================================================================== 

bool
TShortestPath::ProcessNetAccessInfos(TNetAccessInfo *Infos, int NrOfInfos, 
                                     bool SetAlgoInfoTime)

{
    int CInfo;

    for (CInfo = 0; CInfo < NrOfInfos; CInfo++)
    {
        long *SourceTime;
        long *DestTime;

        if ((Infos[CInfo].AlgoInfo = GetAlgoInfo(Infos[CInfo].Node)) == 0)
        {
            LatestError = SP_AlgoInfoNotAvailable;
            //      Warning("ProcessNetAccessInfos: cannot find algo info for %s", 
            //              (char *) Infos[CInfo].Node->GetObjectName());
            return FALSE;
        }
        if (SetAlgoInfoTime)
        {
            SourceTime = &Infos[CInfo].Time;
            DestTime = &Infos[CInfo].AlgoInfoTime;
        }
        else
        {
            Infos[CInfo].AlgoInfoTime = (long) Infos[CInfo].AlgoInfo->GetDist();
            SourceTime = &Infos[CInfo].AlgoInfoTime;        
            DestTime = &Infos[CInfo].Time;
        }
        if (Infos[CInfo].Edge)
        {
            double      TravelTimeOnEdge;
            TParallelID EdgeID;
            
            EdgeID = Infos[CInfo].Edge->GetParallelID();
            if ((TravelTimeOnEdge = GetEdgeTravelTime(Infos[CInfo].Edge, *SourceTime)) == INVALID)
            {
                //              Error("ProcessNetAccessInfos: cannot retrieve travel time on access edge %s",
                //                    (char *) Infos[CInfo].Edge->GetInstanceName());
                LatestError = SP_TravelTimeNotAvailable;
                return FALSE;
            }
            *DestTime = *SourceTime + (long) (Infos[CInfo].FractionOnEdge * TravelTimeOnEdge);
        }
        else
            *DestTime = *SourceTime;
        if (!SetAlgoInfoTime)
            Infos[CInfo].Time += SourceTimeOffset;
    }
    return TRUE;
}

// =================================================================================== 

void
TShortestPath::ResetAlgoInfos(void)

{
    int CInfo;

    for (CInfo = 0; CInfo < Nodes.GetNrOfElements(); CInfo++)
        AlgoInfos[CInfo].Reset();
}

// =================================================================================== 
#undef ParentClass
