/*      file     : TrafficData/TTrafficDataManager.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TTrafficDataManager.C,v 1.18 2005/04/01 15:11:25 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <File.h>
#include <TDebug.h>
#include <TStream.h>
#include <TTransferInfo.h>
#include <TConfiguration.h>
#include <TEnvironment.h>

#include <ParToolsConfig.h>
#include <TLocationInfo.h>
#include <TIncidenceInfo.h>
#include <TAdjacenceInfo.h>
#include <TClassInfo.h>
#include <TNode.h>
#include <TEdge.h>
#include <TObjectManager.h>
#include <TStatManager.h>
#include <TObjectTimer.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>
#include <TObjectTimeStat.h>
#include <TAsciiObjectTimeStatReader.h>

#include "TrafficDataTypes.h"
#include "TrafficDataConfig.h"

#include "TTrafficTimers.h"
#include "TGraphEdgeInfo.h"
#include "TDatabaseNode.h"
#include "TDatabaseEdge.h"
#include "TEdgeInfo.h"
#include "TMultilaneGrid.h"
#include "TIntersectionControl.h"
#include "TAccessory.h"
#include "TAccessoryHandler.h"
#include "TSource.h"
#include "TSink.h"
#include "TAsciiAccessoryReader.h"
#include "TAsciiNodeReader.h"
#include "TAsciiEdgeReader.h"
#include "TAsciiPhaseReader.h"
#include "TTrafficDataManager.h"

// ***DEFINES************************************************************************* 

// ***GLOBAL VARIABLES****************************************************************  

TTrafficDataManager *TrafficDataManager = 0;

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TTrafficDataManager::TTrafficDataManager(void)

{
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));
    if (TrafficDataManager != 0)
    {
        Fail("another instance of TTrafficDataManager already active!");
    }
    TrafficDataManager = this;
    _Init();
}

// ===================================================================================

TTrafficDataManager::~TTrafficDataManager(void)

{
    debug(Print(DM_LIFESPAN, "DESTRUCTOR"));
    Dispose();
    TrafficDataManager = 0;
}

// ===================================================================================

bool
TTrafficDataManager::Activate(void)

{
    debug(Print(DM_LIFESPAN, "Activate"));
    ObjectManager->RegisterClass(0, TIncidenceInfo::ClassInfo, CLASS_NODE, 
                                 CLASS_INCIDENCE, this);
    ObjectManager->RegisterClass(0, TLocationInfo::ClassInfo, 
                                 CLASS_NODE, CLASS_LOCATION, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, NodeTimerClassInfo,
                                 CLASS_NODE, CLASS_NODE_TIMER, this);
    ObjectManager->RegisterClass(0, NodeTimeStatClassInfo,
                                 CLASS_NODE, CLASS_NODE_TIME_STAT, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TAdjacenceInfo::ClassInfo, 
                                 CLASS_EDGE, CLASS_ADJACENCE, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TIntersectionControl::ClassInfo, 
                                 CLASS_NODE, CLASS_INTERSECTION_CONTROL, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TAccessory::ClassInfo, 
                                 0, CLASS_ACCESSORY, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, EdgeTimerClassInfo,
                                 CLASS_EDGE, CLASS_EDGE_TIMER, this);
    ObjectManager->RegisterClass(0, EdgeTimeStatClassInfo,
                                 CLASS_EDGE, CLASS_EDGE_TIME_STAT, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TAccessoryHandler::ClassInfo,
                                 CLASS_EDGE, CLASS_ACCESSORY_HANDLER, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TEdgeInfo::ClassInfo, 
                                 CLASS_EDGE, CLASS_EDGE_INFO, this);
    ObjectManager->RegisterClass(0, TGraphEdgeInfo::ClassInfo, 
                                 CLASS_EDGE, CLASS_GRAPH_EDGE_INFO, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TSource::ClassInfo, 
                                 0, CLASS_SOURCE, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TSink::ClassInfo, 
                                 0, CLASS_SINK, this);
    Configuration->SetDefault(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_NODETIMESTATINTERVAL,
                                           TRAFFIC_DEFAULT_NODETIMESTATINTERVAL));
    Configuration->SetDefault(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_LOADNODETIMES,
                                           TRAFFIC_LOADNODETIMES));
    Configuration->SetDefault(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_EDGETIMESTATINTERVAL,
                                           TRAFFIC_DEFAULT_EDGETIMESTATINTERVAL));
    Configuration->SetDefault(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_LOADEDGETIMES,
                                           TRAFFIC_LOADEDGETIMES));
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::ActivateStatistics(void)

{
    if (!(StatManager->RegisterSamplingHandler(EdgeTimeStatClassInfo->GetID(), this) AND
          StatManager->RegisterSamplingHandler(NodeTimeStatClassInfo->GetID(), this)))
    {
        Error("ActivateStatistics: cannot register myself as sampling handler");
        return FALSE;
    }
    EdgeTimeStatInterval = (int) Configuration->GetLong(TRAFFIC_GROUP, 
                                                        TRAFFIC_NAME_EDGETIMESTATINTERVAL);
    if (EdgeTimeStatInterval > 0)
    {
        TimeEdges = TRUE;
        if (Parallel->IsMaster())
        {
            EdgeTimeSamplingDescription = 
              new TSamplingDescription(EdgeTimeStatClassInfo->GetID(), SF_MERGE,
                                       Parallel->GetLocalCPNID(), SAMPLING_ID_EDGE_TIME_STAT);
            EdgeTimeSamplingDescription->SetCollectInterval(EdgeTimeStatInterval);
            EdgeTimeAggregationInfo = new TAggregationInfo(this, EdgeTimeSamplingDescription);
            StatManager->RegisterAggregation(EdgeTimeAggregationInfo);
            StatManager->StartSampling(EdgeTimeSamplingDescription);
            if (Configuration->Exists(TRAFFIC_GROUP, TRAFFIC_NAME_SAVEEDGETIMES))
            {
                char Filename[512];
                                
                if (Configuration->Exists(TRAFFIC_GROUP, TRAFFIC_NAME_EDGETIMEFILENAME))
                    sprintf(Filename, "%s.%s", Configuration->GetString(), TRAFFIC_NEWEDGETIMEEXTENSION);
                else
                    sprintf(Filename, "%s.%s", BaseFilename(), TRAFFIC_NEWEDGETIMEEXTENSION);
                EdgeTimeStatOutput = new TStream;                   
                if (!EdgeTimeStatOutput->Open(Environment->ReplacePathTokens(Filename), "w"))
                {
                    Error("ActivateStatistics: cannot open edge time stat file %s", Filename);
                    return FALSE;
                }
            }
        }
    }
    NodeTimeStatInterval = (int) Configuration->GetLong(TRAFFIC_GROUP, 
                                                        TRAFFIC_NAME_NODETIMESTATINTERVAL);
    if (NodeTimeStatInterval > 0)
    {
        TimeNodes = TRUE;
        if (Parallel->IsMaster())
        {
            NodeTimeSamplingDescription = 
              new TSamplingDescription(NodeTimeStatClassInfo->GetID(), SF_MERGE,
                                       Parallel->GetLocalCPNID(), SAMPLING_ID_NODE_TIME_STAT);
            NodeTimeSamplingDescription->SetCollectInterval(NodeTimeStatInterval);
            NodeTimeAggregationInfo = new TAggregationInfo(this, NodeTimeSamplingDescription);
            StatManager->RegisterAggregation(NodeTimeAggregationInfo);
            StatManager->StartSampling(NodeTimeSamplingDescription);
            if (Configuration->Exists(TRAFFIC_GROUP, TRAFFIC_NAME_SAVENODETIMES))
            {
                char Filename[512];
                    
                if (Configuration->Exists(TRAFFIC_GROUP, TRAFFIC_NAME_NODETIMEFILENAME))                    
                    sprintf(Filename, "%s.%s", Configuration->GetString(), TRAFFIC_NEWNODETIMEEXTENSION);
                else
                    sprintf(Filename, "%s.%s", BaseFilename(), TRAFFIC_NEWNODETIMEEXTENSION);
                NodeTimeStatOutput = new TStream;                   
                if (!NodeTimeStatOutput->Open(Environment->ReplacePathTokens(Filename), "w"))
                {
                    Error("ActivateStatistics: cannot open node time stat file %s", Filename);
                    return FALSE;
                }
            }
        }
    }
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::CheckDataFiles(long OpenMask)

{
    char    Filename[256];
    TString BaseFilename;

    if (!Configuration->Exists(TRAFFIC_GROUP, TRAFFIC_NAME_BASEFILENAME))
    {
        Error("CheckDataFiles: no base filename set!");
        return FALSE;
    }
    BaseFilename = Environment->GetApplicationFilename(Configuration->GetString());
    if (OpenMask & TRAFFIC_NODES)
    {
        sprintf(Filename, "%s.%s", BaseFilename(), 
                Configuration->GetString(TRAFFIC_GROUP, TRAFFIC_NAME_NODEEXTENSION));
        if (!fexist(Filename))
        {
            Error("CheckDataFiles: no node file");
            return FALSE;
        }
    }
    if (OpenMask & TRAFFIC_EDGES)
    {
        sprintf(Filename, "%s.%s", BaseFilename(), 
                Configuration->GetString(TRAFFIC_GROUP, TRAFFIC_NAME_EDGEEXTENSION));
        if (!fexist(Filename))
        {
            Error("CheckDataFiles: no edge file");
            return FALSE;
        }
    }
    if (OpenMask & TRAFFIC_ACCESSORIES)
    {
        sprintf(Filename, "%s.%s", BaseFilename(), 
                Configuration->GetString(TRAFFIC_GROUP, TRAFFIC_NAME_ACCESSORYEXTENSION));
        if (!fexist(Filename))
        {
            Error("CheckDataFiles: no accessory file");
            return FALSE;
        }
    }
    if (OpenMask & TRAFFIC_PHASES)
    {
        sprintf(Filename, "%s.%s", BaseFilename(), 
                Configuration->GetString(TRAFFIC_GROUP, TRAFFIC_NAME_PHASEEXTENSION));
        if (!fexist(Filename))
        {
            Error("CheckDataFiles: no phase file");
            return FALSE;
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TTrafficDataManager::Collect(TSamplingDescription *Description)

{
    if (Description->GetStatObjectClassID() == EdgeTimeStatClassInfo->GetID())    
        return CollectObjectTimers(Description->GetTimeStep(), Description->GetObjectList(),
                                   EdgeTimerClassInfo, EdgeTimeStatClassInfo);
    if (Description->GetStatObjectClassID() == NodeTimeStatClassInfo->GetID())
        return CollectObjectTimers(Description->GetTimeStep(), Description->GetObjectList(),
                                   NodeTimerClassInfo, NodeTimeStatClassInfo);
    return TStatHandler::Collect(Description);
}

// =================================================================================== 

bool
TTrafficDataManager::CollectObjectTimers(const TTimeStep &TimeStep, TStatObjectList *StatList,
                                         TClassInfo *TimerClass, TClassInfo *TimeStatClass)

{
    TEnumerator< TObjectTimer > CTimer(TimerClass->GetTable());

    CTimer.Enumerate();
    while (CTimer())
    {
        TObjectTimeStat *Stat;

        Stat = new TObjectTimeStat(TimeStatClass, CTimer->GetParallelID());
        Stat->SetTimeStep(TimeStep);
        Stat->Collect(CTimer);
        CTimer->Reset();
        StatList->Append(Stat);
	//        Print("collect %s", (char *) CTimer->GetObjectName());
        CTimer++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TTrafficDataManager::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                             TTimeStep &TimeStep)

{
    if (AggregationInfo == EdgeTimeAggregationInfo)
        return HandleTimeResult(AggregationInfo, TimeStep, EdgeTimeStatOutput);
    if (AggregationInfo == NodeTimeAggregationInfo)
        return HandleTimeResult(AggregationInfo, TimeStep, NodeTimeStatOutput);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 

bool
TTrafficDataManager::HandleTimeResult(TAggregationInfo *AggregationInfo, 
                                      TTimeStep &TimeStep, TStream *Output)

{
    TAggregationResult    *Result = 0;
    TObjectTimeStatList   *StatList = 0;
    TObjectTimeStatEnum    CInfo;
    TTransferInfo          TransferInfo;

    if ((Result = (TAggregationResult *) AggregationInfo->GetResult(TimeStep)) == 0)
    {
        Error("HandleTimeResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((StatList = (TObjectTimeStatList *) Result->StatObjectList) == 0)
    {
        Error("HandleTimeResult: cannot retrieve stat list");
        return FALSE;
    }
    TransferInfo.TransferMode = TM_AsciiData;
    CInfo.WalksForwardThrough(StatList);
    while (CInfo())
    {
        //      Print("write %s", (char *) CInfo->GetObjectName());
        if (!CInfo->Encode(Output, &TransferInfo))
        {
            Error("HandeTimeResult: cannot write to file");
            return FALSE;
        }
        CInfo++;
    }
    AggregationInfo->DisposeResult(TimeStep);
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::LoadEdgeTimeStat(TObjectTimeStat *TimeStat)

{    
    TObjectTimeStat *Stat;
    long             Count;
    int              CStat;

    if (EdgeTimeStatReader == 0)
    {
        TAdjacenceInfo *AdjInfo;

        if ((AdjInfo = (TAdjacenceInfo *) TimeStat->GetInfo(TAdjacenceInfo::ClassInfo->GetInfoID())) == 0)
        {
            Error("LoadEdgeTimeStat: cannot retrieve adjacence info");
            return FALSE;
        }
        //      Print("edge weight=%f", AdjInfo->GetLength(TLocationInfo::ClassInfo->GetInfoID()));

        TimeStat->SetTime(AdjInfo->GetLength(TLocationInfo::ClassInfo->GetInfoID()));
        return TRUE;
    }
    if (!EdgeTimeStatReader->Select(TimeStat->GetParallelID(), Count))
    {
        Error("LoadEdgeTimeStat: cannot select");
        return FALSE;
    }
    for (CStat = 0; CStat < Count; CStat++)
    {
        if ((Stat = (TObjectTimeStat *) EdgeTimeStatReader->GetSelected(CStat)) == 0)
        {
            Error("LoadEdgeTimeStat: cannot retrieve stat");
            return FALSE;
        }
        TimeStat->Accumulate(Stat);
        delete Stat;
    }    
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::LoadNodeTimeStat(TObjectTimeStat *TimeStat)

{
    TObjectTimeStat *Stat;
    long             Count;
    int              CStat;

    if (NodeTimeStatReader == 0)
        return TRUE;

    //    Print("load %s", (char *) TimeStat->GetObjectName());

    if (!NodeTimeStatReader->Select(TimeStat->GetParallelID(), Count))
    {
        Error("LoadNodeTimeStat: cannot select");
        return FALSE;
    }
    for (CStat = 0; CStat < Count; CStat++)
    {
        if ((Stat = (TObjectTimeStat *) NodeTimeStatReader->GetSelected(CStat)) == 0)
        {
            Error("LoadNodeTimeStat: cannot retrieve stat");
            return FALSE;
        }
        TimeStat->Accumulate(Stat);
        delete Stat;
    }    
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::LoadIntersectionControl(TIntersectionControl *Control)

{
    if (PhaseReader == 0)
    {
        Error("CreateIntersectionControl: no phase reader available");
        return 0;
    }
    if (!Control->LoadPhases(PhaseReader))
    {
        Error("cannot assign phases to %s", (char *) Control->GetObjectName());
        return FALSE;
    }
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::CreateSimulationObjects(TNodeList *NodeList, TEdgeList *EdgeList,
                                             TInfoMask NodeLocalMask, 
                                             TInfoMask EdgeLocalMask)

{
    TParallelID *IndexArray;
    long         Count;

    if (TrafficDataManager == 0)
    {
        Error("Distribute: no TTrafficDataManager available");
        return FALSE;
    }
    if (!GetNodeIndexArray(IndexArray, Count))
    {
        Error("cannot retrieve node index array");
        return FALSE;
    }
    Print(DM_CONTROL, "create %ld simulation nodes", Count);
    if (!ObjectManager->AddContainersInIndex(TNode::ClassInfo, IndexArray, Count,
                                             (TContainerObjectList *) NodeList,
                                             NodeLoadableMask, NodeRequiredMask,
                                             NodeLocalMask))
    {
        Error("cannot create all nodes");
        return FALSE;
    }
    if (!GetEdgeIndexArray(IndexArray, Count))
    {
        Error("cannot retrieve edge index array");
        return FALSE;
    }
    Print(DM_CONTROL, "create %ld simulation edges", Count);
    if (!ObjectManager->AddContainersInIndex(TEdge::ClassInfo, IndexArray, Count,
                                             (TContainerObjectList *) EdgeList,
                                             EdgeLoadableMask, EdgeRequiredMask,
                                             EdgeLocalMask))
    {
        Error("cannot create all edges");
        return FALSE;
    }
    return TRUE;
}

// ===================================================================================

void
TTrafficDataManager::Dispose(void)

{
    if (NodeReader)
    {
        NodeReader->Close();
        DeleteObject(NodeReader);
    }
    if (EdgeReader)
    {
        EdgeReader->Close();
        DeleteObject(EdgeReader);
    }
    if (IncomingEdgeReader)
    {
        IncomingEdgeReader->Close();
        DeleteObject(IncomingEdgeReader);
    }
    if (OutgoingEdgeReader)
    {
        OutgoingEdgeReader->Close();
        DeleteObject(OutgoingEdgeReader);
    }
    if (AccessoryReader)
    {
        AccessoryReader->Close();
        DeleteObject(AccessoryReader);
    }
    if (PhaseReader)
    {
        PhaseReader->Close();
        DeleteObject(PhaseReader);
    }
    if (EdgeTimeStatReader)
    {
        EdgeTimeStatReader->Close();
        DeleteObject(EdgeTimeStatReader);
    }
    if (NodeTimeStatReader)
    {
        NodeTimeStatReader->Close();
        DeleteObject(NodeTimeStatReader);
    }
    DeleteObject(EdgeTimeStatOutput);
    DeleteObject(NodeTimeStatOutput);
}

// ===================================================================================

TString
TTrafficDataManager::GetClassName(void) const

{
    return "TTrafficDataManager";
}

// ===================================================================================

TInfoMask
TTrafficDataManager::GetDistributionEdgeLoadableMask(void) const

{
    return DistributionEdgeLoadableMask;
}

// ===================================================================================

TInfoMask
TTrafficDataManager::GetDistributionNodeLoadableMask(void) const

{
    return DistributionNodeLoadableMask;
}

// ===================================================================================

bool
TTrafficDataManager::GetEdgeIndexArray(TParallelID *&IndexArray, long &Count)

{
    if (EdgeReader == 0)
    {
        Error("GetEdgeIndexArray: no edge reader available");
        return FALSE;
    }
    return EdgeReader->GetLinearIndex(IndexArray, Count);
}

// ===================================================================================

bool
TTrafficDataManager::GetNodeIndexArray(TParallelID *&IndexArray, long &Count)

{
    if (NodeReader == 0)
    {
        Error("GetNodeIndexArray: no node reader available");
        return FALSE;
    }
    return NodeReader->GetLinearIndex(IndexArray, Count);
}

// =================================================================================== 

bool
TTrafficDataManager::HandleCreateRequest(TParallelObject *&Object, 
                                         const TClassInfo *ClassInfo, 
                                         TParallelID ParallelID)

{
    if (ClassInfo == TIncidenceInfo::ClassInfo)
    {
        Object = new TIncidenceInfo((TClassInfo *) ClassInfo, ParallelID);
        //      Print("create %s", (char *) Object->GetObjectName());
    }
    else if (ClassInfo == TEdgeInfo::ClassInfo)

        Object = new TEdgeInfo(ParallelID);

    else if (ClassInfo == NodeTimerClassInfo)

        Object = new TObjectTimer(NodeTimerClassInfo, ParallelID);

    else if (ClassInfo == EdgeTimerClassInfo)

        Object = new TObjectTimer(EdgeTimerClassInfo, ParallelID);

    else if (ClassInfo == NodeTimeStatClassInfo)

        Object = new TObjectTimeStat(NodeTimeStatClassInfo, ParallelID);

    else if (ClassInfo == EdgeTimeStatClassInfo)

        Object = new TObjectTimeStat(EdgeTimeStatClassInfo, ParallelID);

    else if (ClassInfo == TLocationInfo::ClassInfo)

        Object = new TLocationInfo((TClassInfo *) ClassInfo, ParallelID);

    else if (ClassInfo == TAdjacenceInfo::ClassInfo)

        Object = new TAdjacenceInfo((TClassInfo *) ClassInfo, ParallelID);

    else if (ClassInfo == TIntersectionControl::ClassInfo)

        Object = new TIntersectionControl(ParallelID);

    else if (ClassInfo == TAccessoryHandler::ClassInfo)

        Object = new TAccessoryHandler(ParallelID);

    else if (ClassInfo == TIntersectionControl::ClassInfo)

        Object = new TIntersectionControl(ParallelID);

    else if (ClassInfo == TAccessory::ClassInfo)

        Object = new TAccessory(ParallelID);

    else if (ClassInfo == TGraphEdgeInfo::ClassInfo)

        Object = new TGraphEdgeInfo(ParallelID);

    return TRUE;
}

// =================================================================================== 

bool
TTrafficDataManager::HandleLoadRequest(TParallelObject *Object, bool &Loaded)

{
    if (Object->GetClassID() == NodeTimerClassInfo->GetID())
    {
        Loaded = TRUE;
    }
    else if (Object->GetClassID() == EdgeTimerClassInfo->GetID())
    {
        Loaded = TRUE;
    }
    else if (Object->GetClassID() == NodeTimeStatClassInfo->GetID())
    {
        Loaded = TRUE;
        return LoadNodeTimeStat((TObjectTimeStat *) Object);
    }
    else if (Object->GetClassID() == EdgeTimeStatClassInfo->GetID())
    {
        Loaded = TRUE;
        return LoadEdgeTimeStat((TObjectTimeStat *) Object);
    }
    else if (Object->GetClassID() == TLocationInfo::ClassInfo->GetID())
    {
        Loaded = TRUE;
        return LoadLocationInfo((TLocationInfo *) Object);
    }
    else if (Object->GetClassID() == TIncidenceInfo::ClassInfo->GetID())
    {
        Loaded = TRUE;
        return LoadIncidenceInfo((TIncidenceInfo *) Object);
    }
    else if (Object->GetClassID() == TAdjacenceInfo::ClassInfo->GetID())
    {
        Loaded = TRUE;
        return LoadAdjacenceInfo((TAdjacenceInfo *) Object);
    }
    else if (Object->GetClassID() == TEdgeInfo::ClassInfo->GetID())
    {
        Loaded = TRUE;
        return LoadEdgeInfo((TEdgeInfo *) Object);
    }
    else if (Object->GetClassID() == TAccessoryHandler::ClassInfo->GetID())
    {
        Loaded = TRUE;
        return LoadAccessoryHandler((TAccessoryHandler *) Object);
    }
    else if (Object->GetClassID() == TIntersectionControl::ClassInfo->GetID())
    {
        Loaded = TRUE;
        return LoadIntersectionControl((TIntersectionControl *) Object);
    }
    else if (Object->GetClassID() == TGraphEdgeInfo::ClassInfo->GetID())
    {
        Loaded = TRUE;
        return ((TGraphEdgeInfo *) Object)->
            LoadLocations(TAdjacenceInfo::ClassInfo->GetInfoID(),
                          TLocationInfo::ClassInfo->GetInfoID());
    }
    return TRUE;
}

// ===================================================================================

void
TTrafficDataManager::_Init(void)

{
    TConfiguration *C = Configuration;

    NodeReader = 0;
    EdgeReader = 0;
    AccessoryReader = 0;
    PhaseReader = 0;
    EdgeTimeStatReader = 0;
    EdgeTimeStatOutput = 0;
    NodeTimeStatReader = 0;
    NodeTimeStatOutput = 0;

    C->Set(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_NODEEXTENSION, TRAFFIC_NODEEXTENSION));
    C->Set(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_EDGEEXTENSION, TRAFFIC_EDGEEXTENSION));
    C->Set(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_ACCESSORYEXTENSION, 
                        TRAFFIC_ACCESSORYEXTENSION));
    C->Set(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_PHASEEXTENSION, TRAFFIC_PHASEEXTENSION));
    C->Set(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_LOADPHASES, TRAFFIC_LOADPHASES));
    C->Set(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_LOADACCESSORIES, TRAFFIC_LOADACCESSORIES));
    C->Set(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_GRANULARITY, TRAFFIC_GRANULARITY));
    C->SetDefault(TConfigEntry(INTERSECTION_GROUP, INTERSECTION_NAME_NONGREEN_PHASE_FACTOR,
                               INTERSECTION_DEFAULT_NONGREEN_PHASE_FACTOR));
    C->SetDefault(TConfigEntry(SOURCE_GROUP, SOURCE_NAME_INSERTIONGAP,
                               SOURCE_DEFAULT_INSERTIONGAP));
    C->SetDefault(TConfigEntry(SOURCE_GROUP, SOURCE_NAME_INSERTIONSLACK,
                               SOURCE_DEFAULT_INSERTIONSLACK));
}

// ===================================================================================

bool
TTrafficDataManager::LoadAccessoryHandler(TAccessoryHandler *AccessoryHandler)

{
    if (AccessoryReader == 0)
    {
        Error("LoadAccessoryHandler: no accessory reader available");
        return FALSE;
    }
    if (!AccessoryHandler->LoadAccessories(TAdjacenceInfo::ClassInfo->GetInfoID(),
                                           AccessoryReader))
    {
        Error("LoadAccessoryHandler: cannot assign accessories");
        return FALSE;
    }
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::LoadAdjacenceInfo(TAdjacenceInfo *AdjacenceInfo)

{
    TDatabaseEdge   *Edge;

    //    Print("load %ld", AdjacenceInfo->GetParallelID());
    if (EdgeReader == 0)
    {
        Error("HandleLoadRequest: cannot load");
        return FALSE;
    }
    if ((Edge = (TDatabaseEdge *) EdgeReader->GetBuffered(AdjacenceInfo->GetParallelID())) == 0)
    {
        Error("cannot load edge adjacence info");
        return FALSE;
    }
    AdjacenceInfo->SetNodeClassID(TNode::ClassInfo->GetID());
    AdjacenceInfo->SetFromNodeID(Edge->FromNodeID);
    AdjacenceInfo->SetToNodeID(Edge->ToNodeID);
    if (Edge->Length != (double) INVALID)
        AdjacenceInfo->SetLength(MeterToLength(Edge->Length));
    if (AdjacenceInfo->GetLength(TLocationInfo::ClassInfo->GetInfoID()) > Granularity)
        AdjacenceInfo->MarkAsDistributable();
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::LoadEdgeInfo(TEdgeInfo *EdgeInfo)

{
    TDatabaseEdge *Edge;

    if (EdgeReader == 0)
    {
        Error("HandleLoadRequest: cannot load");
        return FALSE;
    }
    if ((Edge = (TDatabaseEdge *) EdgeReader->GetBuffered(EdgeInfo->GetParallelID())) == 0)
    {
        Error("cannot load edge info");
        return FALSE;
    }
    *EdgeInfo = *Edge;
    return TRUE;
}

// =================================================================================== 

bool
TTrafficDataManager::LoadIncidenceInfo(TIncidenceInfo *IncidenceInfo)

{
    TParallelID  InEdges[MAX_CONTAINER_DEPS];
    TParallelID  OutEdges[MAX_CONTAINER_DEPS];
    long         NrOfInEdges, NrOfOutEdges;
    int          CEdge;

    if (IncomingEdgeReader == 0 OR OutgoingEdgeReader == 0)
    {
        Error("LoadIncidenceInfo: no reader active");
        return FALSE;
    }
    if (!IncomingEdgeReader->Select(IncidenceInfo->GetParallelID(), NrOfInEdges))
    {
        Error("LoadIncidenceInfo: cannot select incoming edges");
        return FALSE;
    }
    for (CEdge = 0; CEdge < NrOfInEdges; CEdge++)
    {
        TDatabaseEdge *DatabaseEdge;

        if ((DatabaseEdge = (TDatabaseEdge *) IncomingEdgeReader->GetSelected(CEdge)) == 0)
        {
            Error("LoadIncidenceInfo: cannot retrieve edge from file");
            return FALSE;
        }
        InEdges[CEdge] = DatabaseEdge->ID;
    }
    if (!OutgoingEdgeReader->Select(IncidenceInfo->GetParallelID(), NrOfOutEdges))
    {
        Error("LoadIncidenceInfo: cannot select outgoing edges");
        return FALSE;
    }
    for (CEdge = 0; CEdge < NrOfOutEdges; CEdge++)
    {
        TDatabaseEdge *DatabaseEdge;

        if ((DatabaseEdge = (TDatabaseEdge *) OutgoingEdgeReader->GetSelected(CEdge)) == 0)
        {
            Error("LoadIncidenceInfo: cannot retrieve edge from file");
            return FALSE;
        }
        OutEdges[CEdge] = DatabaseEdge->ID;
    }

    if (!IncidenceInfo->AssignEdges(NrOfInEdges, InEdges, NrOfOutEdges, OutEdges))
    {
        Error("LoadIncidenceInfo: cannot assign edges");
        return FALSE;
    }

    //    Print("assign %s", (char *) IncidenceInfo->GetObjectName());

    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::LoadLocationInfo(TLocationInfo *LocationInfo)

{
    TDatabaseNode *Node;

    if (NodeReader == 0)
    {
        Error("HandleLoadRequest: cannot load");
        return FALSE;
    }
    if ((Node = (TDatabaseNode *) NodeReader->GetBuffered(LocationInfo->GetParallelID())) == 0)
    {
        Error("cannot load node location");
        return FALSE;
    }
    LocationInfo->Loc().SetX(MeterToLength(Node->X));
    LocationInfo->Loc().SetY(MeterToLength(Node->Y));
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::Open(void)

{
    long    OpenMask;
    char    Filename[256];

    OpenMask = TRAFFIC_NODES | TRAFFIC_EDGES;
    if (Configuration->GetLong(TRAFFIC_GROUP, TRAFFIC_NAME_LOADACCESSORIES))
        OpenMask |= TRAFFIC_ACCESSORIES;
    if (Configuration->GetLong(TRAFFIC_GROUP, TRAFFIC_NAME_LOADPHASES))
        OpenMask |= TRAFFIC_PHASES;
    if (Configuration->GetLong(TRAFFIC_GROUP, TRAFFIC_NAME_LOADEDGETIMES))
        OpenMask |= TRAFFIC_EDGE_TIMES;
    if (Configuration->GetLong(TRAFFIC_GROUP, TRAFFIC_NAME_LOADNODETIMES))
        OpenMask |= TRAFFIC_NODE_TIMES;
    NodeLoadableMask = NodeRequiredMask = DistributionNodeLoadableMask = 0;
    EdgeLoadableMask = EdgeRequiredMask = DistributionEdgeLoadableMask = 0;
    if (OpenMask & TRAFFIC_NODES)
    {
        NodeRequiredMask |= TLocationInfo::ClassInfo->GetInfoMask() |
                            TIncidenceInfo::ClassInfo->GetInfoMask();
        NodeLoadableMask |= NodeRequiredMask;
        sprintf(Filename, "%s.%s", BaseFilename(), 
                Configuration->GetString(TRAFFIC_GROUP, TRAFFIC_NAME_NODEEXTENSION));
        NodeReader = new TAsciiNodeReader(Filename);
        if (!NodeReader->StartRandomAccess(1.0))
        {
            Error("cannot open node file");
            return FALSE;
        }
    }
    if (OpenMask & TRAFFIC_EDGES)
    {
        EdgeRequiredMask |= (TAdjacenceInfo::ClassInfo->GetInfoMask() |
                             TEdgeInfo::ClassInfo->GetInfoMask());
        if (TMultilaneGrid::ClassInfo)
            EdgeRequiredMask |= TMultilaneGrid::ClassInfo->GetInfoMask();
        EdgeLoadableMask |= EdgeRequiredMask;
        sprintf(Filename, "%s.%s", BaseFilename(), 
                Configuration->GetString(TRAFFIC_GROUP, TRAFFIC_NAME_EDGEEXTENSION));
        EdgeReader = new TAsciiEdgeReader(Filename);
        if (!EdgeReader->StartRandomAccess(1.0))
        {
            Error("cannot open edge file");
            return FALSE;
        }
        IncomingEdgeReader = new TAsciiEdgeReader(Filename, "%*lld %*lld %lld", 
                                                  TRAFFIC_INCOMINGEDGEEXTENSION);
        if (!IncomingEdgeReader->StartRandomAccess(1.0))
        {
            Error("cannot open edge file");
            return FALSE;
        }
        OutgoingEdgeReader = new TAsciiEdgeReader(Filename, "%*lld %lld",
                                                  TRAFFIC_OUTGOINGEDGEEXTENSION);
        if (!OutgoingEdgeReader->StartRandomAccess(1.0))
        {
            Error("cannot open edge file");
            return FALSE;
        }
    }
    if (OpenMask & TRAFFIC_ACCESSORIES)
    {
        EdgeLoadableMask |= TAccessoryHandler::ClassInfo->GetInfoMask();
        sprintf(Filename, "%s.%s", BaseFilename(), 
                Configuration->GetString(TRAFFIC_GROUP, TRAFFIC_NAME_ACCESSORYEXTENSION));
        AccessoryReader = new TAsciiAccessoryReader(Filename);
        if (!AccessoryReader->StartRandomAccess(1.0))
        {
            Error("cannot open accessory file");
            return FALSE;
        }
    }
    if (OpenMask & TRAFFIC_PHASES)
    {
        NodeLoadableMask |= TIntersectionControl::ClassInfo->GetInfoMask();
        sprintf(Filename, "%s.%s", BaseFilename(), 
                Configuration->GetString(TRAFFIC_GROUP, TRAFFIC_NAME_PHASEEXTENSION));
        PhaseReader = new TAsciiPhaseReader(Filename);
        if (!PhaseReader->StartRandomAccess(1.0))
        {
            Error("cannot open phase file");
            return FALSE;
        }
    }
    if (OpenMask & TRAFFIC_NODE_TIMES)
    {   
        TString RealFilename;

        if (Configuration->Exists(TRAFFIC_GROUP, TRAFFIC_NAME_NODETIMEFILENAME))
            sprintf(Filename, "%s.%s", Configuration->GetString(), TRAFFIC_NODETIMEEXTENSION);
        else
            sprintf(Filename, "%s.%s", BaseFilename(), TRAFFIC_NODETIMEEXTENSION);
        RealFilename = Environment->ReplacePathTokens(Filename);
        if (fexist(RealFilename))
        {
            NodeTimeStatReader = new TAsciiObjectTimeStatReader(NodeTimeStatClassInfo, RealFilename);
            if (!NodeTimeStatReader->StartRandomAccess(1.0))
            {
                Error("cannot open node time stat file");
                return FALSE;
            }
            NodeLoadableMask |= NodeTimeStatClassInfo->GetInfoMask();
            DistributionNodeLoadableMask |= NodeTimeStatClassInfo->GetInfoMask();
            Print("opened node time statistics file %s", (char *) RealFilename);
        }
        else
        {
            Warning("cannot find node time statistics file %s -> will ignore node times", 
                    (char *) RealFilename);
        }
    }
    if (OpenMask & TRAFFIC_EDGE_TIMES)
    {
        TString RealFilename;

        if (Configuration->Exists(TRAFFIC_GROUP, TRAFFIC_NAME_EDGETIMEFILENAME))
            sprintf(Filename, "%s.%s", Configuration->GetString(), TRAFFIC_EDGETIMEEXTENSION);
        else
            sprintf(Filename, "%s.%s", BaseFilename(), TRAFFIC_EDGETIMEEXTENSION);
        RealFilename = Environment->ReplacePathTokens(Filename);
        if (fexist(RealFilename))
        {
            EdgeTimeStatReader = new TAsciiObjectTimeStatReader(EdgeTimeStatClassInfo, RealFilename);
            if (!EdgeTimeStatReader->StartRandomAccess(1.0))
            {
                Error("cannot open edge time stat file");
                return FALSE;
            }
            Print("opened edge time statistics file %s", (char *) RealFilename);
        }
        else
        {
            Warning("cannot find edge time statistics file %s -> will use edge lengths", 
                    (char *) RealFilename);
        }
    }
    EdgeLoadableMask |= EdgeTimeStatClassInfo->GetInfoMask();
    DistributionEdgeLoadableMask |= EdgeTimeStatClassInfo->GetInfoMask();
    Print("data base files opened");
    return TRUE;
}

// ===================================================================================

bool
TTrafficDataManager::Run(void)

{
    debug(Print(DM_LIFESPAN, "Run"));    
    if (!Configuration->Exists(TRAFFIC_GROUP, TRAFFIC_NAME_BASEFILENAME))
    {
        Error("Open: no map base filename set!");
        return FALSE;
    }
    BaseFilename = Environment->GetApplicationFilename(Configuration->GetString());
    TSource::InsertionGap = Configuration->GetLong(SOURCE_GROUP, 
                                                   SOURCE_NAME_INSERTIONGAP);
    TSource::InsertionSlack = Configuration->GetLong(SOURCE_GROUP, 
                                                     SOURCE_NAME_INSERTIONSLACK);
    Granularity = 
        MeterToLength(Configuration->GetDouble(TRAFFIC_GROUP, TRAFFIC_NAME_GRANULARITY));
    if (Parallel == 0 OR Parallel->IsMaster() OR ObjectManager->LoadDataLocally())
        if (!Open())
        {
            Error("Run: cannot open files");
            return FALSE;
        }
    if (Parallel)
        if (!ActivateStatistics())
        {
            Error("Run: cannot activate statistics");
            return FALSE;
        }
    return TRUE;
}

// ===================================================================================
#undef ParentClass

