/*      file     : parallel/TBaseMessage.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TBaseMessage.C,v 1.16 2005/03/25 11:06:20 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Error.h>
#include <TString.h>

#include "TDestination.h"
#include "TCommunicationControl.h"
#include "TBaseMessage.h"

// ***CONSTANTS***********************************************************************

char *MessageStateNames[] =

{
    "new",
    "new receive",
    "new send",
    "receiving",
    "sending",
    "received",
    "sent"
};

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#   include "TBaseMessage.i.C"
#endif

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TBaseMessage::TBaseMessage(TCommunicationControl *theControl, TCPNID theSender)

{
    _Init();
    Control = theControl;
    Sender = theSender;
    State = MSG_STATE_NEW_RECEIVE;
}

// =================================================================================== 

TBaseMessage::TBaseMessage(TCommunicationControl* theControl, TMessageID theID,
                           TCPNID theSender, TDestination *theDestination,
                           TMessageTag theTag, TMessageType theType, 
                           TMessageFlags theFlags)

{
    _Init();
    Control = theControl;
    ID = theID;
    Sender = theSender;
    Destination = theDestination;
    Tag = theTag;
    Type = theType;
    Flags = theFlags;
    State = MSG_STATE_NEW_SEND;
}

// =================================================================================== 

TBaseMessage::~TBaseMessage(void)

{
    switch (State)
    {
    case MSG_STATE_RECEIVED:
    case MSG_STATE_SENT:
        break;

    case MSG_STATE_NEW_RECEIVE:
        Warning("~TBaseMessage: receive not started yet");
        break;

    case MSG_STATE_NEW_SEND:
        Warning("~TBaseMessage: send not started yet");
        break;

    case MSG_STATE_RECEIVING:
        Warning("~TBaseMessage: receive in progress");
        break;

    case MSG_STATE_SENDING:
        Warning("~TBaseMessage: send in progress");
        break;
    }
    DeleteObject(Destination);
}

// =================================================================================== 

bool
TBaseMessage::BeginSend(void)

{
    if (State != MSG_STATE_NEW_SEND)
    {
        Error("BeginSend: message not meant to be sent");
        return FALSE;
    }
    State = MSG_STATE_SENDING;
    debug(Print(DM_MSG, "BeginSend"));
    return (Encode(ID) AND
            Encode(Tag) AND
            Encode(Type) AND
            Encode(Flags));
}

// =================================================================================== 

bool
TBaseMessage::BeginReceive(void)

{
    if (State != MSG_STATE_NEW_RECEIVE)
    {
        Error("BeginReceive: message not meant to be received");
        return FALSE;
    }
    State = MSG_STATE_RECEIVING;
    if (!(Decode(ID) AND
          Decode(Tag) AND
          Decode(Type) AND
          Decode(Flags)))
        return FALSE;
    debug(Print(DM_MSG, "BeginReceive"));
    return TRUE;
}

// =================================================================================== 

bool
TBaseMessage::Deselect(void)

{
    debug(Print(DM_MSG, "Deselect"));
    return TRUE;
}

// =================================================================================== 

bool
TBaseMessage::EndReceive(void)

{
    if (!Select())
        return FALSE;
    if (State != MSG_STATE_RECEIVING)
        Warning("EndReceive: message not receiving");
    State = MSG_STATE_RECEIVED;
    debug(Print(DM_MSG, "EndReceive"));
    return TRUE;
}

// =================================================================================== 

bool
TBaseMessage::EndSend(void)

{
    if (State != MSG_STATE_SENDING)
        Warning("EndSend: message not sending");
    if (Flags & MSG_FLAG_EVENT)
    {
        TEventID EventID = NULL_EVENT_ID;
        if (!Encode(EventID))
            return FALSE;
    }
    State = MSG_STATE_SENT;
    debug(Print(DM_MSG, "EndSend"));
    return TRUE;
}

// =================================================================================== 

bool
TBaseMessage::EventPending(TEventID &anEventID)

{
    if (IsFlushed())
        anEventID = NULL_EVENT_ID;
    else
    {
        if (!Select())
        {
            Print("EventPending: Select failed");
            return FALSE;
        }
        if (!Decode(anEventID))
        {
            Error("cannot decode event ID");
            return FALSE;
        }
    }
    if (anEventID == NULL_EVENT_ID)
        debug(Print(DM_EVENT, "no more events pending"));
    else
        debug(Print(DM_EVENT, "decoding event with ID %ld", anEventID));
    return (anEventID != NULL_EVENT_ID);
}

// =================================================================================== 

bool
TBaseMessage::Flush(void)

{
    Flags |= MSG_FLAG_FLUSHED;
    return TRUE;
}

// =================================================================================== 

TString
TBaseMessage::GetClassName(void) const

{
    return "TBaseMessage";
}

// =================================================================================== 

TString
TBaseMessage::GetInstanceName(void) const

{
    char Line[256];
    
    switch(State)
    {
    case MSG_STATE_NEW:
        return "State:new";
        
    case MSG_STATE_NEW_SEND:
    case MSG_STATE_SENDING:
    case MSG_STATE_SENT:
        sprintf(Line, "ID:%ld %s To:%s Tag:%lX Type:%ld Flags:%ld",
                ID, MessageStateNames[State],
                (char *) Destination->GetInstanceName(), Tag, Type, Flags);
        break;

    case MSG_STATE_NEW_RECEIVE:
    case MSG_STATE_RECEIVING:
    case MSG_STATE_RECEIVED:
        sprintf(Line, "ID:%ld %s From:%lX Tag:%lX Type:%ld Flags:%ld",
                ID, MessageStateNames[State], Sender, Tag, Type, Flags);
        break;
    }
    return Line;
}

// =================================================================================== 

void
TBaseMessage::_Init(void)

{
    Control = 0;
    Destination = 0;
    Flags = 0;
    Sender = INVALID_ID;
    ID = INVALID_ID;
    Type = INVALID;
    State = MSG_STATE_NEW;
    Tag = INVALID;
}

// =================================================================================== 

bool
TBaseMessage::Select(void)

{
    debug(Print(DM_MSG, "Select"));
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
