/*      file     : parallel/TDestination.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TDestination.C,v 1.9 2005/03/25 11:06:20 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <string.h>
#include <stdio.h>

#include "TParallel.h"
#include "TCPNInfo.h"
#include "TDestination.h"

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TDestination.i.C"
#endif

// =================================================================================== 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TDestination::TDestination(TCPNID theRecipient)

{
    _Init();
    Recipient = theRecipient;
    switch (Recipient)
    {
    case DEST_MYSELF:
        Recipient = Parallel->GetLocalCPNID();
        Mutable->NrOfRecipients = 1;    
        break;

    case DEST_PARENT:
        Recipient = Parallel->GetParentCPNID();
        Mutable->NrOfRecipients = 1;    
        break;

    case DEST_ALL_BUT_MYSELF:
        Mutable->Flags |= DEST_FLAG_EXCLUDE_MYSELF;
        break;

    case DEST_ALL:
    case DEST_LIST:
        break;

    default:
        Mutable->NrOfRecipients = 1;
    }
}

// =================================================================================== 

TDestination::TDestination(const TCPNInfo *theRecipient)

{
    _Init();
    Recipient = theRecipient->ID;
    Mutable->NrOfRecipients = 1;
}

// =================================================================================== 

TDestination::TDestination(const TDLList<TCPNInfo> *theCPNInfoList, TDestFlags theFlags)

{
    _Init();
    CPNInfoList = theCPNInfoList;
    Recipient = DEST_LIST;
    Mutable->Flags = theFlags;
}

// =================================================================================== 

TDestination::TDestination(const TDestination &aDestination)

{
    _Init();
    Recipient = aDestination.Recipient;
    CPNInfoList = aDestination.CPNInfoList;
    if (aDestination.Mutable)
    {
        Mutable->Flags = aDestination.Mutable->Flags;
        Mutable->NrOfRecipients = aDestination.Mutable->NrOfRecipients;
        if (aDestination.Mutable->Recipients)
        {
            if ((Mutable->Recipients = new TCPNID[Mutable->NrOfRecipients]) == 0)
            {
                Fail("out of heap");
            }
            memcpy(Mutable->Recipients, aDestination.Mutable->Recipients, 
                   Mutable->NrOfRecipients * sizeof(TCPNID));
        }
    }
}

// =================================================================================== 

TDestination::~TDestination(void)

{
    if (Mutable)
    {
        DeleteObjectArray(Mutable->Recipients);
        DeleteObject(Mutable);
    }
}

// =================================================================================== 

bool
TDestination::CheckRecipients(void) const

{
    TCPNInfoEnum        CInfo;
    const TCPNInfoList *List = 0;

    if (Mutable->Recipients == 0)
    {
        switch(Recipient)
        {
        case DEST_LIST:
            List = CPNInfoList;
            break;
            
        case DEST_ALL:
        case DEST_ALL_BUT_MYSELF:
            List = Parallel->GetCPNInfoList();
            break;

        case DEST_NEIGHBORS:
            Error("DEST_NEIGHBORS cannot be resolved by TDestination");
            return FALSE;

        default:
            Mutable->Recipients = new TCPNID[1];
            Mutable->NrOfRecipients = 1;
            Mutable->Recipients[0] = Recipient;
            return TRUE;            
        }
        Mutable->Recipients = new TCPNID[List->GetNrOfElements()];
        CInfo.WalksForwardThrough(List);
        Mutable->NrOfRecipients = 0;
        while(CInfo())
        {
            if (!(((CInfo->ID == Parallel->GetLocalCPNID()) AND
                   (Mutable->Flags & DEST_FLAG_EXCLUDE_MYSELF)) OR
                  ((CInfo->ID == Parallel->GetParentCPNID()) AND
                   (Mutable->Flags & DEST_FLAG_EXCLUDE_PARENT))))
            {
                if (CInfo->ID == Parallel->GetLocalCPNID())
                    Mutable->Flags |= DEST_FLAG_INCLUDES_MYSELF;
                if (CInfo->ID == Parallel->GetParentCPNID())
                    Mutable->Flags |= DEST_FLAG_INCLUDES_PARENT;
                Mutable->Recipients[Mutable->NrOfRecipients++] = CInfo->ID;
            }
            CInfo++;
        }
        if (Mutable->NrOfRecipients == 0)
            Warning("CheckRecipients: no recipients found");
    }
    return TRUE;
}

// =================================================================================== 

TString
TDestination::GetClassName(void) const

{
    return "TDestination";
}

// =================================================================================== 

TString
TDestination::GetInstanceName(void) const

{
    char Line[512];
    char Number[32];
    int  i;

    switch (Recipient)
    {
    case INVALID:
        return "INVALID";

    case DEST_LIST:
        if (!CheckRecipients())
            return "<ERROR>";
        Line[0] = 0;
        for (i = 0; i < Mutable->NrOfRecipients; i++)
        {           
            sprintf(Number, " %lX", Mutable->Recipients[i]);
            strcat(Line, Number);
        }
        break;

    case DEST_ALL:
        return "ALL";

    case DEST_ALL_BUT_MYSELF:
        return "ALL_BUT_MYSELF";

    default:
        sprintf(Line, "%lX", Recipient);
    }
    return Line;
}

// =================================================================================== 

void
TDestination::_Init(void)

{
    Mutable = new TDestMutable;
    Mutable->Flags = 0;
    Mutable->NrOfRecipients = INVALID;
    Mutable->Recipients = 0;
    CPNInfoList = 0;
}

// =================================================================================== 

bool
TDestination::IsRecipient(TCPNID aRecipient) const

{
    int i;

    if (!CheckRecipients())
    {
        Error("IsRecipient: error in CheckRecipients");
        return FALSE;
    }
    for (i = 0; i < Mutable->NrOfRecipients; i++)
        if (Mutable->Recipients[i] == aRecipient)
            return TRUE;
    return FALSE;
}

// =================================================================================== 
#undef ParentClass
