/*      file     : TLocalBalancer.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TLocalBalancer.C,v 1.8 2005/03/25 11:06:49 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <limits.h>

#include <TConfiguration.h>
#include <TParallel.h>
#include <TDestination.h>
#include <TTimeStep.h>
#include <TClassInfo.h>
#include <TStatHandler.h>
#include <TStatManager.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>
#include <TLoadStat.h>

#include <ParStatTypes.h>

#include "ParBalanceConfig.h"
#include "TLocalBalancer.h"

// ***DEFINES************************************************************************* 

#ifndef DBL_MAX
#  define DBL_MAX         1.7976931348623157E+308
#endif

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseBalancer
// =================================================================================== 

TLocalBalancer::TLocalBalancer(void)

{
    LoadAggregationInfo = 0;
    LoadSamplingDescription = 0;
}

// =================================================================================== 

TLocalBalancer::~TLocalBalancer(void)

{
    DeleteObject(LoadAggregationInfo);
    DeleteObject(LoadSamplingDescription);
}

// =================================================================================== 

bool
TLocalBalancer::ExecuteSubTimeStep(TTimeStep &TimeStep)

{
    if (!ParentClass::ExecuteSubTimeStep(TimeStep))
        return FALSE;
    return TRUE;
}

// =================================================================================== 

TString
TLocalBalancer::GetClassName(void) const

{
    return "TLocalBalancer";
}

// =================================================================================== 

bool
TLocalBalancer::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                        TTimeStep &TimeStep)

{
    if (AggregationInfo == LoadAggregationInfo)
        return HandleLoadResult(AggregationInfo, TimeStep);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 

bool
TLocalBalancer::HandleLoadResult(TAggregationInfo *AggregationInfo, TTimeStep &TimeStep)

{
    TAggregationResult *Result = 0;
    TLoadStatList      *StatList = 0;
    TLoadStatEnum       CStat;
    double              MinLoad;
    TLoadStat          *MinStat;
    TLoadStat          *MyStat;

    if ((Result = (TAggregationResult *) AggregationInfo->GetResult(TimeStep)) == 0)
    {
        Error("HandleLoadResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((StatList = (TLoadStatList *) Result->StatObjectList) == 0)
    {
        Error("HandleLoadResult: cannot retrieve stat list");
        return FALSE;
    }
    MyStat = 0;
    MinStat = 0;
    MinLoad = DBL_MAX;
    CStat.WalksForwardThrough(StatList);
    while (CStat())
    {   
        Print(DM_LOAD, "effective time on neighbor %X: %f", CStat->GetParallelID(), 
              CStat->GetAverageEffectiveTime());
        if (CStat->GetParallelID() == Parallel->GetLocalCPNID())
            MyStat = CStat;
        else
            if (CStat->GetAverageEffectiveTime() < MinLoad)
            {
                MinLoad = CStat->GetAverageEffectiveTime();
                MinStat = CStat;
            }
        CStat++;
    }
    if (MyStat != 0 AND MinStat != 0)
        if (!SelectTransfer(TimeStep, MyStat, MinStat))
        {
            Error("HandleLoadResult: cannot select transfer");
            return FALSE;
        }
    AggregationInfo->DisposeResult(TimeStep);
    return TRUE;
}

// =================================================================================== 

bool
TLocalBalancer::Run(void)

{
    if (!ParentClass::Run())
        return FALSE;

    LoadSamplingDescription = 
        new TSamplingDescription(TLoadStat::ClassInfo->GetID(), SF_NEIGHBORS,
                                 DEST_NEIGHBORS, SAMPLING_ID_NEIGHBOR_LOAD_STAT);
    LoadSamplingDescription->SetCollectInterval(5);
    LoadAggregationInfo = new TAggregationInfo(this, LoadSamplingDescription);
    StatManager->RegisterAggregation(LoadAggregationInfo);
    StatManager->StartSampling(LoadSamplingDescription);
    return TRUE;
}

// =================================================================================== 

bool
TLocalBalancer::SelectTransfer(TTimeStep &TimeStep, TLoadStat *MyStat, TLoadStat *MinStat)

{
    if (MyStat->GetAverageEffectiveTime() > MinStat->GetAverageEffectiveTime())
    {
        //      double TransferLoad;
        double MyPerf, MyTime;
        double OtherPerf, OtherTime;

        MyTime = MyStat->GetAverageEffectiveTime();
        MyPerf = MyStat->GetAveragePerf();
        OtherTime = MinStat->GetAverageEffectiveTime();
        OtherPerf =  MinStat->GetAveragePerf();

        // effective load is a topology load:
        // TransferLoad = (MyLoad * OtherPerf - OtherLoad * MyPerf) / 
        //    (MyPerf + OtherPerf);

        // effective load is a time:


        if (!SetTransferInTimeUnits(MinStat->GetParallelID(), TimeStep, 
                                    MyTime, MyPerf, OtherTime, OtherPerf))
        {
            Error("SelectTransfer: cannot set transfer to %lX",
                  MinStat->GetParallelID());
            return FALSE;
        }

    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
