/*      file     : parstat/TStatManager.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TStatManager.C,v 1.9 2005/03/25 11:08:01 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <ParToolsConfig.h>
#include <TBaseMessage.h>
#include <TTopology.h>
#include <TCommunication.h>
#include <TObjectManager.h>
#include <TTimeStep.h>

#include "TSamplingDescription.h"
#include "TAggregationInfo.h"
#include "TAggregationResult.h"
#include "TStatObject.h"
#include "TStatHandler.h"
#include "TStatHandlerInfo.h"
#include "TStatManager.h"

// ***GLOBAL VARIABLES****************************************************************  

TStatManager *StatManager = 0;

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TStatManager::TStatManager(void)

{
    SamplingDescriptions = new TSamplingDescriptionList;
    AggregationInfos = new TAggregationInfoTable(CM_AUTO_DISPOSE);
    Handlers = new THandlerTable(CM_AUTO_DISPOSE);
    StatManager = this;
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));
}

// =================================================================================== 

TStatManager::~TStatManager(void)

{
    debug(Print(DM_LIFESPAN, "DESTRUCTOR"));
    StatManager = 0;
    DeleteObject(Handlers);
    DeleteObject(AggregationInfos);
    DeleteObject(SamplingDescriptions);
}

// =================================================================================== 

bool
TStatManager::Activate(void)

{
    debug(Print(DM_LIFESPAN, "Activate"));
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TSamplingDescription::ClassInfo,
                                 0, CLASS_SAMPLING_DESCRIPTION, this);    
    Communication->RegisterRecipient(MTAG_STAT_MGR, this);
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::BroadcastSamplingDescription(TSamplingDescription *Desc)

{
    TBaseMessage             *Message;
    TSamplingDescriptionList  TempList;

    TempList.Append(Desc);
    if (!(Communication->GetNewMessage(Message, DEST_ALL, MTAG_STAT_MGR, MTYPE_STAT_MGR) AND
          Message->Event(MEVENT_STAT_MGR_SAMPLING_DESCRIPTION) AND
          ObjectManager->SendObjectList((TParallelObjectList *) &TempList, Message) AND
          Communication->Commit(Message)))
    {
        Error("BrodcastSamplingDescription: cannot send %s",
              (char *) Desc->GetObjectName());
        return FALSE;
    }
    Print(DM_STAT_SAMPLE, "broadcast sampling description %s", 
          (char *) Desc->GetInstanceName());
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::BroadcastSamplingDescriptions(void)

{
    TBaseMessage *Message;

    if (!(Communication->GetNewMessage(Message, DEST_ALL, MTAG_STAT_MGR, MTYPE_STAT_MGR) AND
          Message->Event(MEVENT_STAT_MGR_SAMPLING_DESCRIPTION) AND
          ObjectManager->SendObjectList((TParallelObjectList *) SamplingDescriptions,
                                        Message) AND
          Communication->Commit(Message)))
    {
        Error("BrodcastSamplingDescriptions: cannot send list");
        return FALSE;
    }
    Print(DM_STAT_SAMPLE, "broadcast %ld sampling descriptions", 
          SamplingDescriptions->GetNrOfElements());
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::CancelAggregation(TAggregationInfo *AggregationInfo)

{
    AggregationInfos->DeleteByReference(AggregationInfo);
    return FALSE;
}

// =================================================================================== 

bool
TStatManager::Collect(TSamplingDescription *Desc, TTimeStep &TimeStep)

{
    TStatHandler    *Handler;
    TBaseMessage    *Message = 0;
    TParallelID      SampleUniqueID;
    TCPNInfoList    *DestList;

    //    Print("collect %s for %s", (char *) Desc->GetObjectName(),
    //    (char *) TimeStep.GetInstanceName());

    if ((Handler = GetHandler(Desc->GetStatObjectClassID())) == 0)
    {
        Error("Collect: cannot find handler for StatObjectClassID %d",
              Desc->GetStatObjectClassID());
        return FALSE;
    }
    Desc->SetTimeStep(TimeStep);
    if (!Handler->Collect(Desc))
    {
        Error("Collect: while collecting %s", (char *) Desc->GetInstanceName());
        return FALSE;
    }
    SampleUniqueID = Desc->GetParallelID();
    if ((DestList = Desc->GetDestList()))
    {
//      Print("sending stat to %ld dests", DestList->GetNrOfElements());

        if (!Communication->GetNewMessage(Message, DestList, MTAG_STAT_MGR, MTYPE_STAT_MGR))
        {
            Error("Collect: cannot send message to list of destinations");
            return FALSE;
        }
    }
    else
        if (Desc->IsNeighborStat())
        {
            if (!Communication->GetNewMessage(Message, ObjectManager->GetNeighbors(),
                                              MTAG_STAT_MGR, MTYPE_STAT_MGR))
            {
                Error("Collect: cannot send message to neighbors");
                return FALSE;
            }
        }
        else
            if (!Communication->GetNewMessage(Message, Desc->GetDestination(),
                                              MTAG_STAT_MGR, MTYPE_STAT_MGR))
            {
                Error("Collect: cannot send message to %lX", Desc->GetDestination());
                return FALSE;
            }
    if (!(Message->Event(MEVENT_STAT_MGR_STAT) AND
          Message->Encode(SampleUniqueID) AND
          TimeStep.Encode(Message) AND
          ObjectManager->SendObjectList((TParallelObjectList *) Desc->GetObjectList(),
                                        Message, 0) AND
          Communication->Commit(Message)))
    {
        Error("Collect: cannot send stat objects");
        return FALSE;
    }
    debug(Print(DM_STAT_SAMPLE,"send %ld stat object(s) for %s to %lX", 
                Desc->GetObjectList()->GetNrOfElements(), (char *) Desc->GetInstanceName(),
                Desc->GetDestination()));

    // Move the resetting to a seperate loop !!!!!

    if (!Desc->NeedsReset())
        Desc->GetObjectList()->DeleteAll(TRUE);
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::DeleteSamplingDescription(TBaseMessage *Message)

{
    TParallelID           SamplingID;
    TSamplingDescription *SamplingDesc;

    if (!Message->Decode(SamplingID))
        return FALSE;
    if ((SamplingDesc = GetSamplingDescription(SamplingID)) == 0)
        return TRUE;
    DeleteObject(SamplingDesc);
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::ExecuteSubTimeStep(TTimeStep &TimeStep)

{
    TSamplingDescriptionEnum CDesc;

    CDesc.WalksForwardThrough(SamplingDescriptions);
    while (CDesc())
    {   
        if (CDesc->NeedsSampling(TimeStep))
            if (!Sample(CDesc, TimeStep))
            {
                Error("ExecuteTimeStep: cannot sample %s", 
                      (char *) CDesc->GetInstanceName());
                return FALSE;
            }
        if (CDesc->NeedsCollecting(TimeStep))
            if (!Collect(CDesc, TimeStep))
            {
                Error("ExecuteTimeStep: cannot collect %s", 
                      (char *) CDesc->GetInstanceName());
                return FALSE;
            }
        CDesc++;
    }
    if (!ResetPreviouslyCollectedStats(TimeStep))
    {
        Error("ExecuteSubTimeStep: cannot reset stats");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

TAggregationInfo *
TStatManager::GetAggregationInfo(TParallelID SamplingID)

{
    TAggregationInfo *Info;

    if ((Info = AggregationInfos->Find(SamplingID)) == 0)
    {
        Error("GetAggregationInfo: cannot find aggregation info for SamplingID %ld",
              SamplingID);
        return 0;
    }
    return Info;
}

// =================================================================================== 

TString
TStatManager::GetClassName(void) const

{
    return "TStatManager";
}

// =================================================================================== 

TStatHandler *
TStatManager::GetHandler(TClassID StatObjectClassID)

{
    TStatHandlerInfo *Entry;

    if ((Entry = Handlers->Find(StatObjectClassID)) == 0)
    {
        Warning("GetHandler: no handler found for StatObjectClassID %d", 
                StatObjectClassID);
        return 0;
    }
    return Entry->Handler;
}

// =================================================================================== 

TString
TStatManager::GetInstanceName(void) const

{
    char Line[256];

    sprintf(Line, "SD:%ld H:%ld AI:%ld", SamplingDescriptions->GetNrOfElements(),
            Handlers->GetNrOfElements(), AggregationInfos->GetNrOfElements());
    return Line;
}

// =================================================================================== 

TSamplingDescription *
TStatManager::GetSamplingDescription(TParallelID SamplingID)

{
    TSamplingDescription *Desc;

    if ((Desc = (TSamplingDescription *) 
         TSamplingDescription::ClassInfo->GetObjectPtr(SamplingID)) == 0)
    {
        Error("GetSamplingDescription: cannot get description for sampling ID %ld",
              SamplingID);
        return 0;
    }
    return Desc;
}

// =================================================================================== 

bool
TStatManager::HandleCreateRequest(TParallelObject *&Object, 
                                  const TClassInfo *ClassInfo, TParallelID ParallelID)

{
    if (ClassInfo == TSamplingDescription::ClassInfo)

        Object = new TSamplingDescription(ParallelID);

    return TRUE;
}

// =================================================================================== 

bool
TStatManager::HandleEvent(TEventID EventID, TBaseMessage *Message, int &Handled)

{
    switch(EventID)
    {
    case MEVENT_STAT_MGR_SAMPLING_DESCRIPTION:  
        Handled = TRUE;
        return ReceiveSamplingDescriptions(Message);

    case MEVENT_STAT_MGR_DELETE_DESCRIPTION:
        Handled = TRUE;
        return DeleteSamplingDescription(Message);

    case MEVENT_STAT_MGR_STAT:
        Handled = TRUE;
        return ReceiveStatObjects(Message);

    default:
        Error("illegal event ID %ld", EventID);
        return FALSE;
    }
#ifndef COMPLETE_SWITCH_FLOW
    return TRUE;
#endif
}


// =================================================================================== 

bool
TStatManager::MergeStatObjects(TSamplingDescription *Desc,
                               TAggregationResult *Result, TStatObjectList *NewList)

{
    if (Desc->IsMergeStat())
    {
        TStatObjectEnum CStat;
        
        CStat.WalksForwardThrough(NewList);
        while (CStat())
        {
            if (CStat->GetFraction() < 0.99)
            {
                TStatObjectEnum  CPendingStat;
                TStatObject     *MergeStat;
                bool             Found = FALSE;

                if (Result->PendingObjectList == 0)
                    Result->PendingObjectList = new TStatObjectList;            
                CPendingStat.WalksForwardThrough(Result->PendingObjectList);
                while (CPendingStat() AND !Found)
                {
                    if (CPendingStat->GetParallelID() == CStat->GetParallelID())
                        Found = TRUE;
                    else
                        CPendingStat++;
                }
                if (Found)
                {
                    MergeStat = CPendingStat;
                    if (!CStat->Accumulate(CPendingStat))
                    {
                        Error("MergeStatObjects: cannot merge %s with %s",
                              (char *) CStat->GetObjectName(), 
                              (char *) MergeStat->GetObjectName());
                        return FALSE;
                    }
                    /*
                    Print("merge %s and %s", 
                          (char *) CStat->GetObjectName(), 
                          (char *) MergeStat->GetObjectName());
                          */
                    Result->PendingObjectList->Delete(MergeStat);
                    Result->StatObjectList->Append(CStat);
                    delete MergeStat;
                }
                else
                    Result->PendingObjectList->Append(CStat);
            }
            else
                Result->StatObjectList->Append(CStat);
            CStat++;
        }           
    }
    else
        Result->StatObjectList->Append(*NewList);
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::ModifySampling(TSamplingDescription *Description)

{
    Print(DM_STAT_SAMPLE, "modify sampling %s", (char *) Description->GetObjectName());
    if (!BroadcastSamplingDescription(Description))
    {
        Error("ModifySampling: cannot send description");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::ProcessStatObjects(TSamplingDescription *Desc, TAggregationInfo *Info,
                                 TTimeStep &TimeStep, TStatObjectList *StatObjects)

{    
    TAggregationResult *Result;
    int                 NrOfCPN;
    TCPNInfoList       *DestList;

    if (Desc->IsNeighborStat())
        NrOfCPN = (int) ObjectManager->GetNeighbors()->GetNrOfElements();
    else
        if ((DestList = Desc->GetDestList()))
        {
            NrOfCPN = DestList->GetNrOfElements();
            //      Print("number of CPN %d", NrOfCPN);
        }
        else
            NrOfCPN = Topology->GetNrOfCPN();
    if ((Result = Info->GetResult(TimeStep, NrOfCPN)) == 0)
    {
        Error("ProcessStatObjects: cannot get result from %s for time-step %s",
              (char *) Info->GetInstanceName(), (char *) TimeStep.GetInstanceName());
        return 0;
    }
    if (Result->NrOfMissingCPN == 0)
        Warning("ProcessStatObjects: received superfluous stat objects");
    Result->NrOfMissingCPN--;
    if (Desc->IsAggregationStat())
    {
        TStatObjectEnum CObject;

        CObject.WalksForwardThrough(StatObjects);
        while (CObject())
        {
            Result->Fraction += CObject->GetFraction();
            if (Result->StatObject)
            {
                Print(DM_STAT_AGGR, "accumulate %s for time-step %s (missing CPN:%d)",
                      (char *) CObject->GetObjectName(), 
                      (char *) TimeStep.GetInstanceName(), Result->NrOfMissingCPN);
                if (!Result->StatObject->Accumulate(CObject))
                {
                    Error("ProcessStatObjects: cannot accumulate for %s in time-step %s",
                          (char *) Info->GetInstanceName(), 
                          (char *) TimeStep.GetInstanceName());
                    return FALSE;
                }
            }
            else
            {
                Print(DM_STAT_AGGR, "first %s for time-step %s (missing CPN:%ld)",
                      (char *) CObject->GetObjectName(),  
                      (char *) TimeStep.GetInstanceName(),Result->NrOfMissingCPN);
                Result->StatObject = CObject;
                StatObjects->Delete(CObject);
            }
            CObject++;
        }
        StatObjects->DeleteAll(TRUE);
    }
    else
    {
        if (Result->StatObjectList == 0)
        {
            Print(DM_STAT_AGGR, "first %ld object(s) of %s  for time-step %s "
                  "(missing CPN:%ld)", StatObjects->GetNrOfElements(), 
                  (char *) Desc->GetInstanceName(), 
                  (char *) TimeStep.GetInstanceName(), Result->NrOfMissingCPN);
            Result->StatObjectList = new TStatObjectList;
            Result->StatObjectList->Append(*StatObjects);
        }
        else
        {
            Print(DM_STAT_AGGR, "append %ld object(s) of %s for time-step %s "
                  "(missing CPN:%ld)", StatObjects->GetNrOfElements(),
                  (char *) Desc->GetInstanceName(), 
                  (char *) TimeStep.GetInstanceName(), Result->NrOfMissingCPN);
            if (!MergeStatObjects(Desc, Result, StatObjects))
            {
                Error("ProcessStatObjects: while merging stat object lists");
                return FALSE;
            }
        }
    }
    if (Result->NrOfMissingCPN == 0 AND !Result->Handled)
    {
        if (Result->StatObject)
            if (!Result->StatObject->Evaluate())
            {
                Error("ProcessStatObjects: cannot evaluate %s", 
                      (char *) Result->StatObject->GetObjectName());
                return FALSE;
            }
        if (Result->StatObjectList)
        {
            TStatObjectEnum CObject;

            if (Result->PendingObjectList)
                StatObjects->Append(*Result->PendingObjectList);
            CObject.WalksForwardThrough(StatObjects);
            while (CObject())
            {
                if (!CObject->Evaluate())
                {
                    Error("ProcessStatObjects: cannot evaluate %s", 
                          (char *) CObject->GetObjectName());
                    return FALSE;
                }
                CObject++;
            }
        }
        if (!Info->GetHandler()->HandleAggregationResult(Info, TimeStep))
        {
            Error("ProcessStatObjects: while calling aggregation handler");
            return FALSE;
        }
    }
    //    Print("%d CPN missing for %s", Result->NrOfMissingCPN, 
    //    (char *) Desc->GetInstanceName());
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::ReceiveSamplingDescriptions(TBaseMessage *Message)

{
    if (!ObjectManager->ReceiveObjectList((TParallelObjectList *) SamplingDescriptions,
                                           Message, TRUE, 0))
    {
        Error("ReceiveSamplingDescriptions: while receiving list");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::ReceiveStatObjects(TBaseMessage *Message)

{
    TParallelID           SamplingID;
    TTimeStep             TimeStep;
    TAggregationInfo     *AggregationInfo;
    TSamplingDescription *SamplingDesc;
    TStatObjectList       StatObjects;

    if (!(Message->Decode(SamplingID) AND
          TimeStep.Decode(Message)))
    {
        Error("ReceiveStatObjects: cannot receive sampling ID");
        return FALSE;
    }
    if ((AggregationInfo = GetAggregationInfo(SamplingID)) == 0)
    {
        Warning("ReceiveStatObjects: cannot get aggregation info for SamplingID %d",
                SamplingID);
        Message->Flush();       
        return TRUE;
    }
    if ((SamplingDesc = GetSamplingDescription(SamplingID)) == 0)
    {
        Warning("ReceiveStatObjects: cannot get sampling description for ID %d",
                SamplingID);
        Message->Flush();
        return TRUE;
    }
    if (!ObjectManager->ReceiveObjectList((TParallelObjectList *) &StatObjects,
                                          Message, FALSE))
    {
        Error("ReceiveStatObjects: while receiving list");
        return FALSE;
    }
    debug(Print(DM_STAT_SAMPLE,"received %ld stat object(s) for %s to %lX",
                StatObjects.GetNrOfElements(), (char *) SamplingDesc->GetInstanceName(),
                Message->GetSender()));
    return ProcessStatObjects(SamplingDesc, AggregationInfo, TimeStep, &StatObjects);
}

// =================================================================================== 

bool
TStatManager::RegisterAggregation(TAggregationInfo *AggregationInfo)

{
    if (!AggregationInfos->Insert(AggregationInfo, TRUE))
    {
        Error("RegisterAggregation: sample ID already handled");
        return FALSE;
    }
    Print(DM_STAT_AGGR, "register %s", (char *) AggregationInfo->GetObjectName());
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::RegisterSamplingHandler(TClassID StatObjectClassID,
                                      TStatHandler *Handler)

{
    TStatHandlerInfo *Entry;

    if (StatObjectClassID == INVALID)

    {
        Error("RegisterSamplingHandler: invalid StatObjectClassID %d", StatObjectClassID);
        return FALSE;
    }
    Entry = new TStatHandlerInfo(StatObjectClassID, Handler);
    if (!Handlers->Insert(Entry, TRUE))
    {
        DeleteObject(Entry);
        Error("RegisterSamplingHandler: handler for StatObjectClassID %d already registered",
              StatObjectClassID);
        return FALSE;
    }
    Print(DM_STAT_SAMPLE, "register handler for sampling ID %ld", StatObjectClassID);
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::ResetPreviouslyCollectedStats(const TTimeStep &TimeStep)

{
    TSamplingDescriptionEnum CDesc;

    CDesc.WalksForwardThrough(SamplingDescriptions);
    while (CDesc())
    {   
        if (CDesc->NeedsCollecting(TimeStep) AND
            CDesc->NeedsReset())
        {
            TStatObjectEnum CStatObject;

            IfDebug(DM_STAT_SAMPLE, 
                    Print("reset stats for %s", (char *) CDesc->GetObjectName()));
            CStatObject.WalksForwardThrough(CDesc->GetObjectList());
            while (CStatObject())
            {
                CStatObject->Reset();
                CStatObject++;
            }
        }
        CDesc++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::Run(TTopology *theTopology)

{
    debug(Print(DM_LIFESPAN, "Run"));
    Topology = theTopology;
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::Sample(TSamplingDescription *Desc, TTimeStep &TimeStep)

{
    TStatHandler    *Handler;

    if ((Handler = GetHandler(Desc->GetStatObjectClassID())) == 0)
    {
        Error("Sample: cannot find handler for StatObjectClassID %d",
              Desc->GetStatObjectClassID());
        return FALSE;
    }
    Desc->SetTimeStep(TimeStep);
    if (!Handler->Sample(Desc))
    {
        Error("Sample: while sampling %s", (char *) Desc->GetInstanceName());
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::StartSampling(TSamplingDescription *Description, bool Broadcast)

{
    if (Parallel)
    {
        TCPNInfoList *DestList;

        if ((DestList = Description->GetDestList()))
            DestList->Append(Parallel->GetLocalCPNInfo());
    }
    SamplingDescriptions->Append(Description);
    Print(DM_STAT_SAMPLE, "start sampling %s", (char *) Description->GetObjectName());
    if (Broadcast)
        if (!BroadcastSamplingDescription(Description))
        {
            Error("StartSampling: cannot send description");
            return FALSE;
        }
    return TRUE;
}

// =================================================================================== 

bool
TStatManager::StopSampling(TSamplingDescription *Description)

{
    TBaseMessage *Message;
    TParallelID   SamplingID;

    SamplingID = Description->GetParallelID();
    SamplingDescriptions->Delete(Description);
    if (!(Communication->GetNewMessage(Message, DEST_ALL, MTAG_STAT_MGR, 
                                       MTYPE_STAT_MGR) AND
          Message->Event(MEVENT_STAT_MGR_DELETE_DESCRIPTION) AND
          Message->Encode(SamplingID) AND
          Communication->Commit(Message)))
    {
        Error("StopSampling: cannot send");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
