/*      file     : partools/TObjectManager.h
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TObjectManager.h,v 1.18 2005/03/25 11:08:15 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#ifndef TOBJECTMANAGER_H
#define TOBJECTMANAGER_H

// ***INCLUDES************************************************************************ 

#include <TBaseObject.h>
#include <TABTree.h>

#include <TMessageRecipient.h>
#include <TTimeStep.h>

#include "TClassInfo.h"
#include "TParallelObject.h"
#include "TContainerObject.h"
#include "TObjectHandler.h"

// ***FORWARDS************************************************************************

#include <TCPNInfo.forward.h>
#include <TExtendedCPNInfo.forward.h>
#include <TCommunicationLink.forward.h>
#include <TBoundary.forward.h>

// ***DEFINES************************************************************************* 

#define NODE_CLASS "Node"
#define EDGE_CLASS "Edge"

// ***CLASSES************************************************************************* 

/** TObjectManager handles instatiation and migration of objects.
 */

class TObjectManager : public TBaseObject, public TMessageRecipient, public TObjectHandler
    
{
public:
    
    /** Constructor. */
    TObjectManager(void);
    
    /** Destructor..*/
    virtual ~TObjectManager(void);
    
    bool
    Action(TClassInfo *ClassInfo, TParallelObject *Object, TObjectAction Action);

    bool
    Activate(void);
    
    bool
    AddContainersInIndex(TClassInfo *ClassInfo, TParallelID *Index, long Count,
                         TContainerObjectList *List,
                         TInfoMask LoadableMask, TInfoMask RequiredMask,
                         TInfoMask LocalMask = 0);

    bool
    AssignInfo(TContainerObject *Object, const TParallelObject *Info);

    TContainerObject *
    CreateInstance(const TClassInfo *ClassInfo, TParallelID ParallelID,
                   TInfoMask LocalMask = 0, bool CountInstance = TRUE);

    TParallelObject *
    CreateObject(const TClassInfo *ClassInfo, TParallelID ParallelID);

    bool
    DestroyObject(TParallelObject *Object);

    bool
    FillContainer(TContainerObject *ContainerObject, TInfoMask RequiredMask = 0);

    bool
    FillContainers(TContainerObjectList *ContainerList, TInfoMask RequiredMask = 0);
    
    inline TClassInfo *
    GetClassInfo(TClassID ClassID) const;
    
    TClassInfo *
    GetClassInfoByName(const char *Name);

    TContainerObject *
    GetContainer(TContainerReference &Ref, TCPNID CPNID = INVALID);

    inline TCPNInfoList *
    GetNeighbors(void);

    bool
    LoadDataLocally(void) const;

    bool
    LoadObject(TClassInfo *ClassInfo, TParallelObject *theObject, bool &Loaded);

    bool
    Migrate(bool ExitLoop = FALSE);

    bool
    MigrateContainerObjects(const TCPNInfo *Destination, TContainerObjectList *List);

    bool
    ReceiveBoundaries(void);

    TParallelObject *
    ReceiveObject(TBaseStream *aStream, bool UseExisting = FALSE,
                  TTransferInfo *TransferInfo = 0);

    bool
    ReceiveObjectList(TParallelObjectList *ObjectList, TBaseStream *aStream, 
                      bool UseExisting, TTransferInfo *TransferInfo = 0);

    /** Register a class by name and receive a unique ID which must used whenever a new 
      class instance is created. */
    bool
    RegisterClass(int Flags, TClassInfo *&theClassInfo, const char *ContainerClassname,
                  const char *ClassName, TObjectHandler *theObjectHandler = 0);
    bool
    RegisterContainerClass(int Flags, TClassInfo *&ClassInfo,
                           const char *ClassName, TObjectHandler *theObjectHandler);
    
    bool
    Run(void);
    
    bool
    SendBoundaries(TTimeStep &TimeStep, TTransferInfo *TransferInfo);

    bool
    SendObject(const TParallelObject *Object, TBaseStream *aStream, 
               TTransferInfo *TransferInfo = 0) const;

    bool
    SendObjectList(const TParallelObjectList *ObjectList, TBaseStream *aStream, 
                   TTransferInfo *TransferInfo = 0) const;

    inline void
    SetGlobalMigrationMonitor(TCPNID CPNID);

    /** Method called after all classes have been registered. 
      This method optionally synchronizes the class IDs on all CPN. 
      */
    bool
    Start(void);

    bool
    Wait(TCPNID Sender);
    
private:
    
    bool
    ActivateClasses(void);
    
    bool
    AddBoundaryContainer(TContainerObject *Object, TCPNID RemoteCPNID);

    bool
    AddContainer(TContainerObject *Object);

    bool
    AddNewBoundaries(TContainerObject *Object, TCPNID FromCPNID);

    bool
    AddRemoteContainer(TContainerObject *Object);

    bool
    DestroyInstance(TContainerObject *Object);

    bool
    RemoveNonLocalInstances(void);

    bool
    DetermineDependantContainers(TContainerObjectList *Containers,
                                 TContainerObjectList *Dependants);

    virtual TString
    GetClassName(void) const;
    
    TString
    GetInstanceName(void) const;

    virtual bool
    HandleEvent(TEventID EventID, TBaseMessage *Message, int &Handled);

    bool
    HandleEventBoundaries(TBaseMessage *Message, int &Handled);

    bool
    InitializeDataAfterReception(TCPNID FromCPNID);

    bool
    MigrateContainerObject(TBaseStream *Stream, TCPNID Dest, 
                           TContainerObject *Object);

    bool
    ReceiveContainerObject(TBaseStream *Stream);
 
    bool
    ReceiveMigrationInfo(TBaseStream *Stream, bool CheckBoundaries);

    bool
    RemoveBoundaryContainer(TContainerObject *Object, TCPNID CPNID);

    bool
    RemoveContainer(TContainerObject *Object);

    bool
    RemoveOldBoundaries(TCPNID FromCPNID);

    bool
    RemoveRemoteContainer(TContainerObject *Object);

    bool
    SendMigrationInfos(void);

    bool
    SendMigrationInfosToGlobalMonitor(void);

    TABTree< TClassInfo >   ClassInfos;
    TClassInfo            **ClassInfoArray;
    bool                    DataReceived;
    TContainerObjectList    RemoteContainers;
    TContainerObjectList    TransferContainers;
    TContainerObjectList    RequiredContainers;
    TContainerObjectList    DependantContainers;
    TContainerObjectList    ContainersToBeRemoved;
    TContainerObjectList    NewContainerObjects;

    TContainerObjectList    LocalContainerObjects;
    TTimeStep               BoundaryTimeStep;
    TBoundaryList           _IncomingBoundaries[2];
    TBoundaryList          *IncomingBoundaries[2];
    long                    MissingBoundarySets[2];
    TCPNInfoList            TransferDestinations;
    TCPNInfoList            Neighbors;
    TCPNID                  GlobalMigrationMonitor;
    TCPNID                  PreviousCPNIDOfLatestContainer;
    bool                    PreviousObjectExisted;
};

// ***EXTERNALS*********************************************************************** 

extern TObjectManager *ObjectManager;

// ***INLINE METHODS******************************************************************

#ifndef NO_INLINE
#  include "TObjectManager.i.C"
#endif

// ===================================================================================
// Local Variables:
// mode:c++
// End:
#endif
